# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from pyhocon import ConfigFactory

from databuilder.transformer.timestamp_string_to_epoch import (
    FIELD_NAME, TIMESTAMP_FORMAT, TimestampStringToEpoch,
)


class TestTimestampStrToEpoch(unittest.TestCase):

    def test_conversion(self) -> None:

        transformer = TimestampStringToEpoch()
        config = ConfigFactory.from_dict({
            FIELD_NAME: 'foo',
        })
        transformer.init(conf=config)

        actual = transformer.transform({'foo': '2020-02-19T19:52:33.1Z'})
        self.assertDictEqual({'foo': 1582141953}, actual)

    def test_conversion_with_format(self) -> None:

        transformer = TimestampStringToEpoch()
        config = ConfigFactory.from_dict({
            FIELD_NAME: 'foo',
            TIMESTAMP_FORMAT: '%Y-%m-%dT%H:%M:%SZ'
        })
        transformer.init(conf=config)

        actual = transformer.transform({'foo': '2020-02-19T19:52:33Z'})
        self.assertDictEqual({'foo': 1582141953}, actual)

    def test_invalid_timestamp(self) -> None:
        transformer = TimestampStringToEpoch()
        config = ConfigFactory.from_dict({
            FIELD_NAME: 'foo',
        })
        transformer.init(conf=config)
        actual = transformer.transform({'foo': '165de33266d4'})
        self.assertEqual(actual['foo'], 0)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from typing import Any

from mock import patch
from pyhocon import ConfigFactory

from databuilder import Scoped
from databuilder.extractor.neo4j_extractor import Neo4jExtractor
from databuilder.models.table_elasticsearch_document import TableESDocument


class TestNeo4jExtractor(unittest.TestCase):

    def setUp(self) -> None:
        config_dict = {
            f'extractor.neo4j.{Neo4jExtractor.GRAPH_URL_CONFIG_KEY}': 'TEST_GRAPH_URL',
            f'extractor.neo4j.{Neo4jExtractor.CYPHER_QUERY_CONFIG_KEY}': 'TEST_QUERY',
            f'extractor.neo4j.{Neo4jExtractor.NEO4J_AUTH_USER}': 'TEST_USER',
            f'extractor.neo4j.{Neo4jExtractor.NEO4J_AUTH_PW}': 'TEST_PW'
        }

        self.conf = ConfigFactory.from_dict(config_dict)

    def text_extraction_with_empty_query_result(self: Any) -> None:
        """
        Test Extraction with empty results from query
        """
        with patch.object(Neo4jExtractor, '_get_driver'):
            extractor = Neo4jExtractor()
            extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                                  scope=extractor.get_scope()))

            extractor.results = ['']
            result = extractor.extract()
            self.assertIsNone(result)

    def test_extraction_with_single_query_result(self: Any) -> None:
        """
        Test Extraction with single result from query
        """
        with patch.object(Neo4jExtractor, '_get_driver'):
            extractor = Neo4jExtractor()
            extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                                  scope=extractor.get_scope()))

            extractor.results = ['test_result']
            result = extractor.extract()
            self.assertEqual(result, 'test_result')

            # Ensure second result is None
            result = extractor.extract()
            self.assertIsNone(result)

    def test_extraction_with_multiple_query_result(self: Any) -> None:
        """
        Test Extraction with multiple result from query
        """
        with patch.object(Neo4jExtractor, '_get_driver'):
            extractor = Neo4jExtractor()
            extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                                  scope=extractor.get_scope()))

            extractor.results = ['test_result1', 'test_result2', 'test_result3']

            result = extractor.extract()
            self.assertEqual(result, 'test_result1')

            result = extractor.extract()
            self.assertEqual(result, 'test_result2')

            result = extractor.extract()
            self.assertEqual(result, 'test_result3')

            # Ensure next result is None
            result = extractor.extract()
            self.assertIsNone(result)

    def test_extraction_with_model_class(self: Any) -> None:
        """
        Test Extraction using model class
        """
        config_dict = {
            f'extractor.neo4j.{Neo4jExtractor.GRAPH_URL_CONFIG_KEY}': 'TEST_GRAPH_URL',
            f'extractor.neo4j.{Neo4jExtractor.CYPHER_QUERY_CONFIG_KEY}': 'TEST_QUERY',
            f'extractor.neo4j.{Neo4jExtractor.NEO4J_AUTH_USER}': 'TEST_USER',
            f'extractor.neo4j.{Neo4jExtractor.NEO4J_AUTH_PW}': 'TEST_PW',
            f'extractor.neo4j.{Neo4jExtractor.MODEL_CLASS_CONFIG_KEY}':
                'databuilder.models.table_elasticsearch_document.TableESDocument'
        }

        self.conf = ConfigFactory.from_dict(config_dict)

        with patch.object(Neo4jExtractor, '_get_driver'):
            extractor = Neo4jExtractor()
            extractor.init(Scoped.get_scoped_conf(conf=self.conf,
                                                  scope=extractor.get_scope()))

            result_dict = dict(database='test_database',
                               cluster='test_cluster',
                               schema='test_schema',
                               name='test_table_name',
                               display_name='test_schema.test_table_name',
                               key='test_table_key',
                               description='test_table_description',
                               last_updated_timestamp=123456789,
                               column_names=['test_col1', 'test_col2', 'test_col3'],
                               column_descriptions=['test_description1', 'test_description2', ''],
                               total_usage=100,
                               unique_usage=5,
                               tags=['hive'],
                               badges=['badge1'],
                               schema_description='schema_description',
                               programmatic_descriptions=['TEST'])

            extractor.results = [result_dict]
            result_obj = extractor.extract()

            self.assertIsInstance(result_obj, TableESDocument)
            self.assertDictEqual(vars(result_obj), result_dict)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from typing import no_type_check
from unittest.mock import ANY

from databuilder.models.table_column_usage import ColumnReader, TableColumnUsage
from databuilder.serializers import (
    mysql_serializer, neo4_serializer, neptune_serializer,
)
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)


class TestTableColumnUsage(unittest.TestCase):

    def setUp(self) -> None:
        col_readers = [
            ColumnReader(
                database='db',
                cluster='gold',
                schema='scm',
                table='foo',
                column='*',
                user_email='john@example.com'
            )
        ]
        self.table_col_usage = TableColumnUsage(col_readers=col_readers)

    @no_type_check  # mypy is somehow complaining on assignment on expected dict.
    def test_serialize(self) -> None:
        node_row = self.table_col_usage.next_node()
        actual = []
        while node_row:
            actual.append(neo4_serializer.serialize_node(node_row))
            node_row = self.table_col_usage.next_node()

        expected = [{'first_name': '',
                     'last_name': '',
                     'full_name': '',
                     'employee_type': '',
                     'is_active:UNQUOTED': True,
                     'updated_at:UNQUOTED': 0,
                     'LABEL': 'User',
                     'slack_id': '',
                     'KEY': 'john@example.com',
                     'github_username': '',
                     'team_name': '',
                     'email': 'john@example.com',
                     'role_name': ''}]
        self.assertEqual(expected, actual)

        rel_row = self.table_col_usage.next_relation()
        actual = []
        while rel_row:
            actual.append(neo4_serializer.serialize_relationship(rel_row))
            rel_row = self.table_col_usage.next_relation()

        expected = [{'read_count:UNQUOTED': 1, 'END_KEY': 'john@example.com', 'START_LABEL': 'Table',
                     'END_LABEL': 'User', 'START_KEY': 'db://gold.scm/foo', 'TYPE': 'READ_BY', 'REVERSE_TYPE': 'READ'}]
        self.assertEqual(expected, actual)

    def test_neptune_serialize(self) -> None:
        rel_row = self.table_col_usage.next_relation()
        actual = []
        while rel_row:
            actual.append(neptune_serializer.convert_relationship(rel_row))
            rel_row = self.table_col_usage.next_relation()
        expected = [[
            {
                NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                    from_vertex_id='Table:db://gold.scm/foo',
                    to_vertex_id='User:john@example.com',
                    label='READ_BY'
                ),
                METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                    from_vertex_id='Table:db://gold.scm/foo',
                    to_vertex_id='User:john@example.com',
                    label='READ_BY'
                ),
                NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Table:db://gold.scm/foo',
                NEPTUNE_RELATIONSHIP_HEADER_TO: 'User:john@example.com',
                NEPTUNE_HEADER_LABEL: 'READ_BY',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'read_count:Long(single)': 1
            },
            {
                NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                    from_vertex_id='User:john@example.com',
                    to_vertex_id='Table:db://gold.scm/foo',
                    label='READ'
                ),
                METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                    from_vertex_id='User:john@example.com',
                    to_vertex_id='Table:db://gold.scm/foo',
                    label='READ'
                ),
                NEPTUNE_RELATIONSHIP_HEADER_FROM: 'User:john@example.com',
                NEPTUNE_RELATIONSHIP_HEADER_TO: 'Table:db://gold.scm/foo',
                NEPTUNE_HEADER_LABEL: 'READ',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'read_count:Long(single)': 1
            }
        ]]
        self.maxDiff = None
        self.assertListEqual(expected, actual)

    def test_mysql_serialize(self) -> None:
        col_readers = [ColumnReader(database='db', cluster='gold', schema='scm', table='foo', column='*',
                                    user_email='john@example.com')]
        table_col_usage = TableColumnUsage(col_readers=col_readers)

        actual = []
        record = table_col_usage.next_record()
        while record:
            actual.append(mysql_serializer.serialize_record(record))
            record = table_col_usage.next_record()

        expected_user = {'rk': 'john@example.com',
                         'first_name': '',
                         'last_name': '',
                         'full_name': '',
                         'employee_type': '',
                         'is_active': True,
                         'updated_at': 0,
                         'slack_id': '',
                         'github_username': '',
                         'team_name': '',
                         'email': 'john@example.com',
                         'role_name': ''}
        expected_usage = {'table_rk': 'db://gold.scm/foo',
                          'user_rk': 'john@example.com',
                          'read_count': 1}
        expected = [expected_user, expected_usage]

        self.assertEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from typing import Any

from pyhocon import ConfigFactory

from databuilder.transformer.regex_str_replace_transformer import (
    ATTRIBUTE_NAME, REGEX_REPLACE_TUPLE_LIST, RegexStrReplaceTransformer,
)


class TestRegexReplacement(unittest.TestCase):

    def test(self) -> None:
        transformer = self._default_test_transformer()

        foo = Foo('abc')
        actual = transformer.transform(foo)

        self.assertEqual('bba', actual.val)

    def test_numeric_val(self) -> None:
        transformer = self._default_test_transformer()

        foo = Foo(6)
        actual = transformer.transform(foo)

        self.assertEqual(6, actual.val)

    def test_none_val(self) -> None:
        transformer = self._default_test_transformer()

        foo = Foo(None)
        actual = transformer.transform(foo)

        self.assertEqual(None, actual.val)

    def _default_test_transformer(self) -> RegexStrReplaceTransformer:
        config = ConfigFactory.from_dict({
            REGEX_REPLACE_TUPLE_LIST: [('a', 'b'), ('c', 'a')],
            ATTRIBUTE_NAME: 'val'
        })

        transformer = RegexStrReplaceTransformer()
        transformer.init(config)

        return transformer

    def test_dict_replace(self) -> None:
        config = ConfigFactory.from_dict({
            REGEX_REPLACE_TUPLE_LIST: [('\\', '\\\\')],
            ATTRIBUTE_NAME: 'val'
        })

        transformer = RegexStrReplaceTransformer()
        transformer.init(config)

        d = {'val': '\\'}

        actual = transformer.transform(d)

        self.assertEqual({'val': '\\\\'}, actual)


class Foo(object):
    def __init__(self, val: Any) -> None:
        self.val = val


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from mock import patch

from databuilder.rest_api.base_rest_api_query import EmptyRestApiQuerySeed, RestApiQuerySeed
from databuilder.rest_api.rest_api_query import RestApiQuery


class TestRestApiQuery(unittest.TestCase):

    def test_rest_api_query_seed(self) -> None:
        rest_api_query = RestApiQuerySeed(seed_record=[
            {'foo': 'bar'},
            {'john': 'doe'}
        ])

        result = [v for v in rest_api_query.execute()]
        expected = [
            {'foo': 'bar'},
            {'john': 'doe'}
        ]

        self.assertListEqual(expected, result)

    def test_empty_rest_api_query_seed(self) -> None:
        rest_api_query = EmptyRestApiQuerySeed()

        result = [v for v in rest_api_query.execute()]
        assert len(result) == 1

    def test_rest_api_query(self) -> None:

        seed_record = [{'foo1': 'bar1'},
                       {'foo2': 'bar2'}]
        seed_query = RestApiQuerySeed(seed_record=seed_record)

        with patch('databuilder.rest_api.rest_api_query.requests.get') as mock_get:
            json_path = 'foo.name'
            field_names = ['name_field']

            mock_get.return_value.json.side_effect = [
                {'foo': {'name': 'john'}},
                {'foo': {'name': 'doe'}},
            ]
            query = RestApiQuery(query_to_join=seed_query, url='foobar', params={},
                                 json_path=json_path, field_names=field_names)

            expected = [
                {'name_field': 'john', 'foo1': 'bar1'},
                {'name_field': 'doe', 'foo2': 'bar2'}
            ]

            for actual in query.execute():
                self.assertDictEqual(expected.pop(0), actual)

    def test_rest_api_query_multiple_fields(self) -> None:

        seed_record = [{'foo1': 'bar1'},
                       {'foo2': 'bar2'}]
        seed_query = RestApiQuerySeed(seed_record=seed_record)

        with patch('databuilder.rest_api.rest_api_query.requests.get') as mock_get:
            json_path = 'foo.[name,hobby]'
            field_names = ['name_field', 'hobby']

            mock_get.return_value.json.side_effect = [
                {'foo': {'name': 'john', 'hobby': 'skiing'}},
                {'foo': {'name': 'doe', 'hobby': 'snowboarding'}},
            ]
            query = RestApiQuery(query_to_join=seed_query, url='foobar', params={},
                                 json_path=json_path, field_names=field_names)

            expected = [
                {'name_field': 'john', 'hobby': 'skiing', 'foo1': 'bar1'},
                {'name_field': 'doe', 'hobby': 'snowboarding', 'foo2': 'bar2'}
            ]

            for actual in query.execute():
                self.assertDictEqual(expected.pop(0), actual)

    def test_compute_subresult_single_field(self) -> None:
        sub_records = RestApiQuery._compute_sub_records(result_list=['1', '2', '3'], field_names=['foo'])

        expected_records = [
            ['1'], ['2'], ['3']
        ]

        assert expected_records == sub_records

        sub_records = RestApiQuery._compute_sub_records(result_list=['1', '2', '3'], field_names=['foo'],
                                                        json_path_contains_or=True)

        assert expected_records == sub_records

    def test_compute_subresult_multiple_fields_json_path_and_expression(self) -> None:
        sub_records = RestApiQuery._compute_sub_records(
            result_list=['1', 'a', '2', 'b', '3', 'c'], field_names=['foo', 'bar'])

        expected_records = [
            ['1', 'a'], ['2', 'b'], ['3', 'c']
        ]

        assert expected_records == sub_records

        sub_records = RestApiQuery._compute_sub_records(
            result_list=['1', 'a', 'x', '2', 'b', 'y', '3', 'c', 'z'], field_names=['foo', 'bar', 'baz'])

        expected_records = [
            ['1', 'a', 'x'], ['2', 'b', 'y'], ['3', 'c', 'z']
        ]

        assert expected_records == sub_records

    def test_compute_subresult_multiple_fields_json_path_or_expression(self) -> None:
        sub_records = RestApiQuery._compute_sub_records(
            result_list=['1', '2', '3', 'a', 'b', 'c'],
            field_names=['foo', 'bar'],
            json_path_contains_or=True
        )

        expected_records = [
            ['1', 'a'], ['2', 'b'], ['3', 'c']
        ]

        self.assertEqual(expected_records, sub_records)

        sub_records = RestApiQuery._compute_sub_records(
            result_list=['1', '2', '3', 'a', 'b', 'c', 'x', 'y', 'z'],
            field_names=['foo', 'bar', 'baz'],
            json_path_contains_or=True)

        expected_records = [
            ['1', 'a', 'x'], ['2', 'b', 'y'], ['3', 'c', 'z']
        ]

        self.assertEqual(expected_records, sub_records)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import ANY

from databuilder.models.graph_serializable import (
    NODE_KEY, NODE_LABEL, RELATION_END_KEY, RELATION_END_LABEL, RELATION_REVERSE_TYPE, RELATION_START_KEY,
    RELATION_START_LABEL, RELATION_TYPE,
)
from databuilder.models.table_owner import TableOwner
from databuilder.models.user import User
from databuilder.serializers import (
    mysql_serializer, neo4_serializer, neptune_serializer,
)
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)

db = 'hive'
SCHEMA = 'BASE'
TABLE = 'TEST'
CLUSTER = 'DEFAULT'
owner1 = 'user1@1'
owner2 = 'user2@2'


class TestTableOwner(unittest.TestCase):

    def setUp(self) -> None:
        super(TestTableOwner, self).setUp()
        self.table_owner = TableOwner(db_name='hive',
                                      schema=SCHEMA,
                                      table_name=TABLE,
                                      cluster=CLUSTER,
                                      owners="user1@1, user2@2 ")

    def test_get_owner_model_key(self) -> None:
        owner = self.table_owner.get_owner_model_key(owner1)
        self.assertEqual(owner, owner1)

    def test_get_metadata_model_key(self) -> None:
        metadata = self.table_owner.get_metadata_model_key()
        self.assertEqual(metadata, 'hive://DEFAULT.BASE/TEST')

    def test_create_nodes(self) -> None:
        expected_node1 = {
            NODE_KEY: User.USER_NODE_KEY_FORMAT.format(email=owner1),
            NODE_LABEL: User.USER_NODE_LABEL,
            User.USER_NODE_EMAIL: owner1
        }
        expected_node2 = {
            NODE_KEY: User.USER_NODE_KEY_FORMAT.format(email=owner2),
            NODE_LABEL: User.USER_NODE_LABEL,
            User.USER_NODE_EMAIL: owner2
        }
        expected = [expected_node1, expected_node2]

        actual = []
        node = self.table_owner.create_next_node()
        while node:
            serialized_node = neo4_serializer.serialize_node(node)
            actual.append(serialized_node)
            node = self.table_owner.create_next_node()

        self.assertEqual(actual, expected)

    def test_create_nodes_neptune(self) -> None:
        expected_node1 = {
            NEPTUNE_HEADER_ID: "User:" + User.USER_NODE_KEY_FORMAT.format(email=owner1),
            METADATA_KEY_PROPERTY_NAME: "User:" + User.USER_NODE_KEY_FORMAT.format(email=owner1),
            NEPTUNE_HEADER_LABEL: User.USER_NODE_LABEL,
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
            User.USER_NODE_EMAIL + ":String(single)": owner1
        }
        expected_node2 = {
            NEPTUNE_HEADER_ID: "User:" + User.USER_NODE_KEY_FORMAT.format(email=owner2),
            METADATA_KEY_PROPERTY_NAME: "User:" + User.USER_NODE_KEY_FORMAT.format(email=owner2),
            NEPTUNE_HEADER_LABEL: User.USER_NODE_LABEL,
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
            User.USER_NODE_EMAIL + ":String(single)": owner2
        }
        expected = [expected_node1, expected_node2]

        actual = []
        node = self.table_owner.create_next_node()
        while node:
            serialized_node = neptune_serializer.convert_node(node)
            actual.append(serialized_node)
            node = self.table_owner.create_next_node()

        self.assertEqual(actual, expected)

    def test_create_relation(self) -> None:
        expected_relation1 = {
            RELATION_START_KEY: owner1,
            RELATION_START_LABEL: User.USER_NODE_LABEL,
            RELATION_END_KEY: self.table_owner.get_metadata_model_key(),
            RELATION_END_LABEL: 'Table',
            RELATION_TYPE: TableOwner.OWNER_TABLE_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableOwner.TABLE_OWNER_RELATION_TYPE
        }
        expected_relation2 = {
            RELATION_START_KEY: owner2,
            RELATION_START_LABEL: User.USER_NODE_LABEL,
            RELATION_END_KEY: self.table_owner.get_metadata_model_key(),
            RELATION_END_LABEL: 'Table',
            RELATION_TYPE: TableOwner.OWNER_TABLE_RELATION_TYPE,
            RELATION_REVERSE_TYPE: TableOwner.TABLE_OWNER_RELATION_TYPE
        }
        expected = [expected_relation1, expected_relation2]

        actual = []
        relation = self.table_owner.create_next_relation()
        while relation:
            serialized_relation = neo4_serializer.serialize_relationship(relation)
            actual.append(serialized_relation)
            relation = self.table_owner.create_next_relation()

        self.assertEqual(actual, expected)

    def test_create_relation_neptune(self) -> None:
        expected = [
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="User:" + owner1,
                        to_vertex_id="Table:" + self.table_owner.get_metadata_model_key(),
                        label=TableOwner.OWNER_TABLE_RELATION_TYPE
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="User:" + owner1,
                        to_vertex_id="Table:" + self.table_owner.get_metadata_model_key(),
                        label=TableOwner.OWNER_TABLE_RELATION_TYPE
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: "User:" + owner1,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: "Table:" + self.table_owner.get_metadata_model_key(),
                    NEPTUNE_HEADER_LABEL: TableOwner.OWNER_TABLE_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Table:" + self.table_owner.get_metadata_model_key(),
                        to_vertex_id="User:" + owner1,
                        label=TableOwner.TABLE_OWNER_RELATION_TYPE
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Table:" + self.table_owner.get_metadata_model_key(),
                        to_vertex_id="User:" + owner1,
                        label=TableOwner.TABLE_OWNER_RELATION_TYPE
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: "Table:" + self.table_owner.get_metadata_model_key(),
                    NEPTUNE_RELATIONSHIP_HEADER_TO: "User:" + owner1,
                    NEPTUNE_HEADER_LABEL: TableOwner.TABLE_OWNER_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ],
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="User:" + owner2,
                        to_vertex_id="Table:" + self.table_owner.get_metadata_model_key(),
                        label=TableOwner.OWNER_TABLE_RELATION_TYPE
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="User:" + owner2,
                        to_vertex_id="Table:" + self.table_owner.get_metadata_model_key(),
                        label=TableOwner.OWNER_TABLE_RELATION_TYPE
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: "User:" + owner2,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: "Table:" + self.table_owner.get_metadata_model_key(),
                    NEPTUNE_HEADER_LABEL: TableOwner.OWNER_TABLE_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Table:" + self.table_owner.get_metadata_model_key(),
                        to_vertex_id="User:" + owner2,
                        label=TableOwner.TABLE_OWNER_RELATION_TYPE
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Table:" + self.table_owner.get_metadata_model_key(),
                        to_vertex_id="User:" + owner2,
                        label=TableOwner.TABLE_OWNER_RELATION_TYPE
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: "Table:" + self.table_owner.get_metadata_model_key(),
                    NEPTUNE_RELATIONSHIP_HEADER_TO: "User:" + owner2,
                    NEPTUNE_HEADER_LABEL: TableOwner.TABLE_OWNER_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ]
        ]

        actual = []
        relation = self.table_owner.create_next_relation()
        while relation:
            serialized_relation = neptune_serializer.convert_relationship(relation)
            actual.append(serialized_relation)
            relation = self.table_owner.create_next_relation()

        self.assertEqual(expected, actual)

    def test_create_records(self) -> None:
        expected = [
            {
                'rk': User.USER_NODE_KEY_FORMAT.format(email=owner1),
                'email': owner1
            },
            {
                'table_rk': self.table_owner.get_metadata_model_key(),
                'user_rk': owner1
            },
            {
                'rk': User.USER_NODE_KEY_FORMAT.format(email=owner2),
                'email': owner2
            },
            {
                'table_rk': self.table_owner.get_metadata_model_key(),
                'user_rk': owner2
            }
        ]

        actual = []
        record = self.table_owner.create_next_record()
        while record:
            serialized_record = mysql_serializer.serialize_record(record)
            actual.append(serialized_record)
            record = self.table_owner.create_next_record()

        self.assertEqual(actual, expected)

    def test_create_nodes_with_owners_list(self) -> None:
        self.table_owner_list = TableOwner(db_name='hive',
                                           schema=SCHEMA,
                                           table_name=TABLE,
                                           cluster=CLUSTER,
                                           owners=['user1@1', ' user2@2 '])
        expected_node1 = {
            NODE_KEY: User.USER_NODE_KEY_FORMAT.format(email=owner1),
            NODE_LABEL: User.USER_NODE_LABEL,
            User.USER_NODE_EMAIL: owner1
        }
        expected_node2 = {
            NODE_KEY: User.USER_NODE_KEY_FORMAT.format(email=owner2),
            NODE_LABEL: User.USER_NODE_LABEL,
            User.USER_NODE_EMAIL: owner2
        }
        expected = [expected_node1, expected_node2]

        actual = []
        node = self.table_owner_list.create_next_node()
        while node:
            serialized_node = neo4_serializer.serialize_node(node)
            actual.append(serialized_node)
            node = self.table_owner_list.create_next_node()

        self.assertEqual(actual, expected)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import unittest
import uuid

from mock import MagicMock, patch
from neo4j import GraphDatabase
from pyhocon import ConfigFactory

from databuilder.publisher import neo4j_csv_publisher
from databuilder.publisher.neo4j_csv_publisher import Neo4jCsvPublisher

here = os.path.dirname(__file__)


class TestPublish(unittest.TestCase):

    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)
        self._resource_path = os.path.join(here, f'../resources/csv_publisher')

    def test_publisher(self) -> None:
        with patch.object(GraphDatabase, 'driver') as mock_driver:
            mock_session = MagicMock()
            mock_driver.return_value.session.return_value = mock_session

            mock_transaction = MagicMock()
            mock_session.begin_transaction.return_value = mock_transaction

            mock_run = MagicMock()
            mock_transaction.run = mock_run
            mock_commit = MagicMock()
            mock_transaction.commit = mock_commit

            publisher = Neo4jCsvPublisher()

            conf = ConfigFactory.from_dict(
                {neo4j_csv_publisher.NEO4J_END_POINT_KEY: 'dummy://999.999.999.999:7687/',
                 neo4j_csv_publisher.NODE_FILES_DIR: f'{self._resource_path}/nodes',
                 neo4j_csv_publisher.RELATION_FILES_DIR: f'{self._resource_path}/relations',
                 neo4j_csv_publisher.NEO4J_USER: 'neo4j_user',
                 neo4j_csv_publisher.NEO4J_PASSWORD: 'neo4j_password',
                 neo4j_csv_publisher.JOB_PUBLISH_TAG: str(uuid.uuid4())}
            )
            publisher.init(conf)
            publisher.publish()

            self.assertEqual(mock_run.call_count, 6)

            # 2 node files, 1 relation file
            self.assertEqual(mock_commit.call_count, 1)

    def test_preprocessor(self) -> None:
        with patch.object(GraphDatabase, 'driver') as mock_driver:
            mock_session = MagicMock()
            mock_driver.return_value.session.return_value = mock_session

            mock_transaction = MagicMock()
            mock_session.begin_transaction.return_value = mock_transaction

            mock_run = MagicMock()
            mock_transaction.run = mock_run
            mock_commit = MagicMock()
            mock_transaction.commit = mock_commit

            mock_preprocessor = MagicMock()
            mock_preprocessor.is_perform_preprocess.return_value = MagicMock(return_value=True)
            mock_preprocessor.preprocess_cypher.return_value = ('MATCH (f:Foo) RETURN f', {})

            publisher = Neo4jCsvPublisher()

            conf = ConfigFactory.from_dict(
                {neo4j_csv_publisher.NEO4J_END_POINT_KEY: 'dummy://999.999.999.999:7687/',
                 neo4j_csv_publisher.NODE_FILES_DIR: f'{self._resource_path}/nodes',
                 neo4j_csv_publisher.RELATION_FILES_DIR: f'{self._resource_path}/relations',
                 neo4j_csv_publisher.RELATION_PREPROCESSOR: mock_preprocessor,
                 neo4j_csv_publisher.NEO4J_USER: 'neo4j_user',
                 neo4j_csv_publisher.NEO4J_PASSWORD: 'neo4j_password',
                 neo4j_csv_publisher.JOB_PUBLISH_TAG: str(uuid.uuid4())}
            )
            publisher.init(conf)
            publisher.publish()

            self.assertEqual(mock_run.call_count, 8)

            # 2 node files, 1 relation file
            self.assertEqual(mock_commit.call_count, 1)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import copy
import unittest
from typing import Dict, List
from unittest.mock import ANY

from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.serializers import (
    mysql_serializer, neo4_serializer, neptune_serializer,
)
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)


class TestDashboardMetadata(unittest.TestCase):
    def setUp(self) -> None:
        self.full_dashboard_metadata = DashboardMetadata(
            'Product - Jobs.cz',
            'Agent',
            'Agent dashboard description',
            ['test_tag', 'tag2'],
            dashboard_group_description='foo dashboard group description',
            created_timestamp=123456789,
            dashboard_group_url='https://foo.bar/dashboard_group/foo',
            dashboard_url='https://foo.bar/dashboard_group/foo/dashboard/bar'
        )

        # Without tags
        self.dashboard_metadata2 = DashboardMetadata('Product - Atmoskop',
                                                     'Atmoskop',
                                                     'Atmoskop dashboard description',
                                                     [],
                                                     )

        # One common tag with dashboard_metadata, no description
        self.dashboard_metadata3 = DashboardMetadata('Product - Jobs.cz',
                                                     'Dohazovac',
                                                     '',
                                                     ['test_tag', 'tag3']
                                                     )

        # Necessary minimum -- NOT USED
        self.dashboard_metadata4 = DashboardMetadata('',
                                                     'PzR',
                                                     '',
                                                     []
                                                     )

        self.expected_nodes_deduped = [
            {
                'KEY': '_dashboard://gold',
                'LABEL': 'Cluster', 'name': 'gold'
            },
            {
                'created_timestamp:UNQUOTED': 123456789,
                'name': 'Agent',
                'KEY': '_dashboard://gold.Product - Jobs.cz/Agent',
                'LABEL': 'Dashboard',
                'dashboard_url': 'https://foo.bar/dashboard_group/foo/dashboard/bar'
            },
            {'name': 'Product - Jobs.cz', 'KEY': '_dashboard://gold.Product - Jobs.cz', 'LABEL': 'Dashboardgroup',
             'dashboard_group_url': 'https://foo.bar/dashboard_group/foo'},
            {'KEY': '_dashboard://gold.Product - Jobs.cz/_description', 'LABEL': 'Description',
             'description': 'foo dashboard group description'},
            {'description': 'Agent dashboard description',
             'KEY': '_dashboard://gold.Product - Jobs.cz/Agent/_description', 'LABEL': 'Description'},
            {'tag_type': 'dashboard', 'KEY': 'test_tag', 'LABEL': 'Tag'},
            {'tag_type': 'dashboard', 'KEY': 'tag2', 'LABEL': 'Tag'}
        ]

        self.expected_nodes = copy.deepcopy(self.expected_nodes_deduped)

        self.expected_rels_deduped = [
            {'END_KEY': '_dashboard://gold.Product - Jobs.cz', 'END_LABEL': 'Dashboardgroup',
             'REVERSE_TYPE': 'DASHBOARD_GROUP_OF', 'START_KEY': '_dashboard://gold',
             'START_LABEL': 'Cluster', 'TYPE': 'DASHBOARD_GROUP'},
            {'END_KEY': '_dashboard://gold.Product - Jobs.cz/_description', 'END_LABEL': 'Description',
             'REVERSE_TYPE': 'DESCRIPTION_OF', 'START_KEY': '_dashboard://gold.Product - Jobs.cz',
             'START_LABEL': 'Dashboardgroup', 'TYPE': 'DESCRIPTION'},
            {'END_KEY': '_dashboard://gold.Product - Jobs.cz', 'START_LABEL': 'Dashboard',
             'END_LABEL': 'Dashboardgroup',
             'START_KEY': '_dashboard://gold.Product - Jobs.cz/Agent', 'TYPE': 'DASHBOARD_OF',
             'REVERSE_TYPE': 'DASHBOARD'},
            {'END_KEY': '_dashboard://gold.Product - Jobs.cz/Agent/_description', 'START_LABEL': 'Dashboard',
             'END_LABEL': 'Description',
             'START_KEY': '_dashboard://gold.Product - Jobs.cz/Agent', 'TYPE': 'DESCRIPTION',
             'REVERSE_TYPE': 'DESCRIPTION_OF'},
            {'END_KEY': 'test_tag', 'START_LABEL': 'Dashboard', 'END_LABEL': 'Tag',
             'START_KEY': '_dashboard://gold.Product - Jobs.cz/Agent', 'TYPE': 'TAG', 'REVERSE_TYPE': 'TAG_OF'},
            {'END_KEY': 'tag2', 'START_LABEL': 'Dashboard', 'END_LABEL': 'Tag',
             'START_KEY': '_dashboard://gold.Product - Jobs.cz/Agent', 'TYPE': 'TAG', 'REVERSE_TYPE': 'TAG_OF'}
        ]

        self.expected_rels = copy.deepcopy(self.expected_rels_deduped)

        self.expected_nodes_deduped2 = [
            {'KEY': '_dashboard://gold', 'LABEL': 'Cluster', 'name': 'gold'},
            {'name': 'Atmoskop', 'KEY': '_dashboard://gold.Product - Atmoskop/Atmoskop', 'LABEL': 'Dashboard'},
            {'name': 'Product - Atmoskop', 'KEY': '_dashboard://gold.Product - Atmoskop', 'LABEL': 'Dashboardgroup'},
            {'description': 'Atmoskop dashboard description',
             'KEY': '_dashboard://gold.Product - Atmoskop/Atmoskop/_description',
             'LABEL': 'Description'},
        ]

        self.expected_nodes2 = copy.deepcopy(self.expected_nodes_deduped2)

        self.expected_rels_deduped2 = [
            {'END_KEY': '_dashboard://gold.Product - Atmoskop', 'END_LABEL': 'Dashboardgroup',
             'REVERSE_TYPE': 'DASHBOARD_GROUP_OF', 'START_KEY': '_dashboard://gold',
             'START_LABEL': 'Cluster', 'TYPE': 'DASHBOARD_GROUP'},
            {'END_KEY': '_dashboard://gold.Product - Atmoskop', 'START_LABEL': 'Dashboard',
             'END_LABEL': 'Dashboardgroup',
             'START_KEY': '_dashboard://gold.Product - Atmoskop/Atmoskop', 'TYPE': 'DASHBOARD_OF',
             'REVERSE_TYPE': 'DASHBOARD'},
            {'END_KEY': '_dashboard://gold.Product - Atmoskop/Atmoskop/_description', 'START_LABEL': 'Dashboard',
             'END_LABEL': 'Description',
             'START_KEY': '_dashboard://gold.Product - Atmoskop/Atmoskop', 'TYPE': 'DESCRIPTION',
             'REVERSE_TYPE': 'DESCRIPTION_OF'},
        ]

        self.expected_rels2 = copy.deepcopy(self.expected_rels_deduped2)

        self.expected_nodes_deduped3 = [
            {'KEY': '_dashboard://gold', 'LABEL': 'Cluster', 'name': 'gold'},
            {'name': 'Dohazovac', 'KEY': '_dashboard://gold.Product - Jobs.cz/Dohazovac', 'LABEL': 'Dashboard'},
            {'name': 'Product - Jobs.cz', 'KEY': '_dashboard://gold.Product - Jobs.cz', 'LABEL': 'Dashboardgroup'},
            {'tag_type': 'dashboard', 'KEY': 'test_tag', 'LABEL': 'Tag'},
            {'tag_type': 'dashboard', 'KEY': 'tag3', 'LABEL': 'Tag'}
        ]

        self.expected_nodes3 = copy.deepcopy(self.expected_nodes_deduped3)

        self.expected_rels_deduped3 = [
            {'END_KEY': '_dashboard://gold.Product - Jobs.cz', 'END_LABEL': 'Dashboardgroup',
             'REVERSE_TYPE': 'DASHBOARD_GROUP_OF', 'START_KEY': '_dashboard://gold',
             'START_LABEL': 'Cluster', 'TYPE': 'DASHBOARD_GROUP'},
            {'END_KEY': '_dashboard://gold.Product - Jobs.cz', 'START_LABEL': 'Dashboard',
             'END_LABEL': 'Dashboardgroup',
             'START_KEY': '_dashboard://gold.Product - Jobs.cz/Dohazovac', 'TYPE': 'DASHBOARD_OF',
             'REVERSE_TYPE': 'DASHBOARD'},
            {'END_KEY': 'test_tag', 'START_LABEL': 'Dashboard', 'END_LABEL': 'Tag',
             'START_KEY': '_dashboard://gold.Product - Jobs.cz/Dohazovac', 'TYPE': 'TAG', 'REVERSE_TYPE': 'TAG_OF'},
            {'END_KEY': 'tag3', 'START_LABEL': 'Dashboard', 'END_LABEL': 'Tag',
             'START_KEY': '_dashboard://gold.Product - Jobs.cz/Dohazovac', 'TYPE': 'TAG', 'REVERSE_TYPE': 'TAG_OF'},
        ]

        self.expected_rels3 = copy.deepcopy(self.expected_rels_deduped3)

    def test_full_example(self) -> None:
        node_row = self.full_dashboard_metadata.next_node()
        actual = []
        while node_row:
            node_serialized = neo4_serializer.serialize_node(node_row)
            actual.append(node_serialized)
            node_row = self.full_dashboard_metadata.next_node()

        self.assertEqual(self.expected_nodes, actual)

        relation_row = self.full_dashboard_metadata.next_relation()
        actual = []
        while relation_row:
            relation_serialized = neo4_serializer.serialize_relationship(relation_row)
            actual.append(relation_serialized)
            relation_row = self.full_dashboard_metadata.next_relation()

        self.assertEqual(self.expected_rels, actual)

    def test_full_dashboard_example_neptune(self) -> None:
        expected_neptune_rels = [
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Cluster:_dashboard://gold',
                        to_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        label='DASHBOARD_GROUP'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Cluster:_dashboard://gold',
                        to_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        label='DASHBOARD_GROUP'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Cluster:_dashboard://gold',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                    NEPTUNE_HEADER_LABEL: 'DASHBOARD_GROUP',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        to_vertex_id='Cluster:_dashboard://gold',
                        label='DASHBOARD_GROUP_OF'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        to_vertex_id='Cluster:_dashboard://gold',
                        label='DASHBOARD_GROUP_OF'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Cluster:_dashboard://gold',
                    NEPTUNE_HEADER_LABEL: 'DASHBOARD_GROUP_OF',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ],
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        to_vertex_id='Description:_dashboard://gold.Product - Jobs.cz/_description',
                        label='DESCRIPTION'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        to_vertex_id='Description:_dashboard://gold.Product - Jobs.cz/_description',
                        label='DESCRIPTION'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Description:_dashboard://gold.Product - Jobs.cz/_description',
                    NEPTUNE_HEADER_LABEL: 'DESCRIPTION',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Description:_dashboard://gold.Product - Jobs.cz/_description',
                        to_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        label='DESCRIPTION_OF'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Description:_dashboard://gold.Product - Jobs.cz/_description',
                        to_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        label='DESCRIPTION_OF'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Description:_dashboard://gold.Product - Jobs.cz/_description',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                    NEPTUNE_HEADER_LABEL: 'DESCRIPTION_OF',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ],
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        to_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        label='DASHBOARD_OF'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        to_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        label='DASHBOARD_OF'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                    NEPTUNE_HEADER_LABEL: 'DASHBOARD_OF',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        to_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        label='DASHBOARD'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                        to_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        label='DASHBOARD'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                    NEPTUNE_HEADER_LABEL: 'DASHBOARD',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ],
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        to_vertex_id='Description:_dashboard://gold.Product - Jobs.cz/Agent/_description',
                        label='DESCRIPTION'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        to_vertex_id='Description:_dashboard://gold.Product - Jobs.cz/Agent/_description',
                        label='DESCRIPTION'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                    NEPTUNE_RELATIONSHIP_HEADER_TO:
                        'Description:_dashboard://gold.Product - Jobs.cz/Agent/_description',
                    NEPTUNE_HEADER_LABEL: 'DESCRIPTION',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Description:_dashboard://gold.Product - Jobs.cz/Agent/_description',
                        to_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        label='DESCRIPTION_OF'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Description:_dashboard://gold.Product - Jobs.cz/Agent/_description',
                        to_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        label='DESCRIPTION_OF'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM:
                        'Description:_dashboard://gold.Product - Jobs.cz/Agent/_description',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                    NEPTUNE_HEADER_LABEL: 'DESCRIPTION_OF',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ],
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        to_vertex_id='Tag:test_tag',
                        label='TAG'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        to_vertex_id='Tag:test_tag',
                        label='TAG'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Tag:test_tag',
                    NEPTUNE_HEADER_LABEL: 'TAG',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Tag:test_tag',
                        to_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        label='TAG_OF'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Tag:test_tag',
                        to_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        label='TAG_OF'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Tag:test_tag',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                    NEPTUNE_HEADER_LABEL: 'TAG_OF',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ],
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        to_vertex_id='Tag:tag2',
                        label='TAG'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        to_vertex_id='Tag:tag2',
                        label='TAG'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Tag:tag2',
                    NEPTUNE_HEADER_LABEL: 'TAG',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Tag:tag2',
                        to_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        label='TAG_OF'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Tag:tag2',
                        to_vertex_id='Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                        label='TAG_OF'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Tag:tag2',
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                    NEPTUNE_HEADER_LABEL: 'TAG_OF',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ],
        ]

        expected_neptune_nodes = [
            {
                NEPTUNE_HEADER_ID: 'Cluster:_dashboard://gold',
                METADATA_KEY_PROPERTY_NAME: 'Cluster:_dashboard://gold',
                NEPTUNE_HEADER_LABEL: 'Cluster',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'name:String(single)': 'gold'
            },
            {
                NEPTUNE_HEADER_ID: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                METADATA_KEY_PROPERTY_NAME: 'Dashboard:_dashboard://gold.Product - Jobs.cz/Agent',
                NEPTUNE_HEADER_LABEL: 'Dashboard',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'name:String(single)': 'Agent',
                'dashboard_url:String(single)': 'https://foo.bar/dashboard_group/foo/dashboard/bar',
                'created_timestamp:Long(single)': 123456789,
            },
            {
                NEPTUNE_HEADER_ID: 'Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                METADATA_KEY_PROPERTY_NAME: 'Dashboardgroup:_dashboard://gold.Product - Jobs.cz',
                NEPTUNE_HEADER_LABEL: 'Dashboardgroup',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'name:String(single)': 'Product - Jobs.cz',
                'dashboard_group_url:String(single)': 'https://foo.bar/dashboard_group/foo'
            },
            {
                NEPTUNE_HEADER_ID: 'Description:_dashboard://gold.Product - Jobs.cz/_description',
                METADATA_KEY_PROPERTY_NAME: 'Description:_dashboard://gold.Product - Jobs.cz/_description',
                NEPTUNE_HEADER_LABEL: 'Description',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'description:String(single)': 'foo dashboard group description',
            },
            {
                NEPTUNE_HEADER_ID: 'Description:_dashboard://gold.Product - Jobs.cz/Agent/_description',
                METADATA_KEY_PROPERTY_NAME: 'Description:_dashboard://gold.Product - Jobs.cz/Agent/_description',
                NEPTUNE_HEADER_LABEL: 'Description',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'description:String(single)': 'Agent dashboard description'
            },
            {
                NEPTUNE_HEADER_ID: 'Tag:test_tag',
                METADATA_KEY_PROPERTY_NAME: 'Tag:test_tag',
                NEPTUNE_HEADER_LABEL: 'Tag',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'tag_type:String(single)': 'dashboard'
            },
            {
                NEPTUNE_HEADER_ID: 'Tag:tag2',
                METADATA_KEY_PROPERTY_NAME: 'Tag:tag2',
                NEPTUNE_HEADER_LABEL: 'Tag',
                NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
                'tag_type:String(single)': 'dashboard'
            },
        ]
        self.maxDiff = None
        node_row = self.full_dashboard_metadata.next_node()
        actual = []
        while node_row:
            node_serialized = neptune_serializer.convert_node(node_row)
            actual.append(node_serialized)
            node_row = self.full_dashboard_metadata.next_node()

        self.assertEqual(expected_neptune_nodes, actual)

        relation_row = self.full_dashboard_metadata.next_relation()
        neptune_actual: List[List[Dict]] = []
        while relation_row:
            relation_serialized = neptune_serializer.convert_relationship(relation_row)
            neptune_actual.append(relation_serialized)
            relation_row = self.full_dashboard_metadata.next_relation()

        self.assertEqual(expected_neptune_rels, neptune_actual)

    def test_dashboard_without_tags(self) -> None:
        node_row = self.dashboard_metadata2.next_node()
        actual = []
        while node_row:
            node_serialized = neo4_serializer.serialize_node(node_row)
            actual.append(node_serialized)
            node_row = self.dashboard_metadata2.next_node()

        self.assertEqual(self.expected_nodes_deduped2, actual)

        relation_row = self.dashboard_metadata2.next_relation()
        actual = []
        while relation_row:
            relation_serialized = neo4_serializer.serialize_relationship(relation_row)
            actual.append(relation_serialized)
            relation_row = self.dashboard_metadata2.next_relation()

        self.assertEqual(self.expected_rels_deduped2, actual)

    def test_dashboard_no_description(self) -> None:
        node_row = self.dashboard_metadata3.next_node()
        actual = []
        while node_row:
            node_serialized = neo4_serializer.serialize_node(node_row)
            actual.append(node_serialized)
            node_row = self.dashboard_metadata3.next_node()

        self.assertEqual(self.expected_nodes_deduped3, actual)

        relation_row = self.dashboard_metadata3.next_relation()
        actual = []
        while relation_row:
            relation_serialized = neo4_serializer.serialize_relationship(relation_row)
            actual.append(relation_serialized)
            relation_row = self.dashboard_metadata3.next_relation()

        self.assertEqual(self.expected_rels_deduped3, actual)

    def test_dashboard_record_full_example(self) -> None:
        expected_records = [
            {'rk': '_dashboard://gold', 'name': 'gold'},
            {'rk': '_dashboard://gold.Product - Jobs.cz', 'name': 'Product - Jobs.cz',
             'cluster_rk': '_dashboard://gold',
             'dashboard_group_url': 'https://foo.bar/dashboard_group/foo'},
            {'rk': '_dashboard://gold.Product - Jobs.cz/_description', 'description': 'foo dashboard group description',
             'dashboard_group_rk': '_dashboard://gold.Product - Jobs.cz'},
            {'rk': '_dashboard://gold.Product - Jobs.cz/Agent', 'name': 'Agent',
             'dashboard_group_rk': '_dashboard://gold.Product - Jobs.cz', 'created_timestamp': 123456789,
             'dashboard_url': 'https://foo.bar/dashboard_group/foo/dashboard/bar'},
            {'rk': '_dashboard://gold.Product - Jobs.cz/Agent/_description',
             'description': 'Agent dashboard description',
             'dashboard_rk': '_dashboard://gold.Product - Jobs.cz/Agent'},
            {'rk': 'test_tag', 'tag_type': 'dashboard'},
            {'dashboard_rk': '_dashboard://gold.Product - Jobs.cz/Agent', 'tag_rk': 'test_tag'},
            {'rk': 'tag2', 'tag_type': 'dashboard'},
            {'dashboard_rk': '_dashboard://gold.Product - Jobs.cz/Agent', 'tag_rk': 'tag2'}
        ]
        record = self.full_dashboard_metadata.next_record()
        actual = []
        while record:
            record_serialized = mysql_serializer.serialize_record(record)
            actual.append(record_serialized)
            record = self.full_dashboard_metadata.next_record()

        self.assertEqual(expected_records, actual)

    def test_dashboard_record_without_tags(self) -> None:
        expected_records_without_tags = [
            {'rk': '_dashboard://gold', 'name': 'gold'},
            {'rk': '_dashboard://gold.Product - Atmoskop', 'name': 'Product - Atmoskop',
             'cluster_rk': '_dashboard://gold'},
            {'rk': '_dashboard://gold.Product - Atmoskop/Atmoskop', 'name': 'Atmoskop',
             'dashboard_group_rk': '_dashboard://gold.Product - Atmoskop'},
            {'rk': '_dashboard://gold.Product - Atmoskop/Atmoskop/_description',
             'description': 'Atmoskop dashboard description',
             'dashboard_rk': '_dashboard://gold.Product - Atmoskop/Atmoskop'}
        ]
        record = self.dashboard_metadata2.next_record()
        actual = []
        while record:
            record_serialized = mysql_serializer.serialize_record(record)
            actual.append(record_serialized)
            record = self.dashboard_metadata2.next_record()

        self.assertEqual(expected_records_without_tags, actual)

    def test_dashboard_record_no_description(self) -> None:
        expected_records_without_description = [
            {'rk': '_dashboard://gold', 'name': 'gold'},
            {'rk': '_dashboard://gold.Product - Jobs.cz', 'name': 'Product - Jobs.cz',
             'cluster_rk': '_dashboard://gold'},
            {'rk': '_dashboard://gold.Product - Jobs.cz/Dohazovac', 'name': 'Dohazovac',
             'dashboard_group_rk': '_dashboard://gold.Product - Jobs.cz'},
            {'rk': 'test_tag', 'tag_type': 'dashboard'},
            {'dashboard_rk': '_dashboard://gold.Product - Jobs.cz/Dohazovac', 'tag_rk': 'test_tag'},
            {'rk': 'tag3', 'tag_type': 'dashboard'},
            {'dashboard_rk': '_dashboard://gold.Product - Jobs.cz/Dohazovac', 'tag_rk': 'tag3'}
        ]
        record = self.dashboard_metadata3.next_record()
        actual = []
        while record:
            record_serialized = mysql_serializer.serialize_record(record)
            actual.append(record_serialized)
            record = self.dashboard_metadata3.next_record()

        self.assertEqual(expected_records_without_description, actual)


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import shutil
import tempfile
import unittest
from typing import Any

from mock import patch
from pyhocon import ConfigFactory, ConfigTree

from databuilder.extractor.base_extractor import Extractor
from databuilder.job.job import DefaultJob
from databuilder.loader.base_loader import Loader
from databuilder.task.task import DefaultTask
from databuilder.transformer.base_transformer import Transformer

LOGGER = logging.getLogger(__name__)


class TestJob(unittest.TestCase):

    def setUp(self) -> None:
        self.temp_dir_path = tempfile.mkdtemp()
        self.dest_file_name = f'{self.temp_dir_path}/superhero.json'
        self.conf = ConfigFactory.from_dict({'loader.superhero.dest_file': self.dest_file_name})

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir_path)

    def test_job(self) -> None:
        with patch("databuilder.job.job.StatsClient") as mock_statsd:
            task = DefaultTask(SuperHeroExtractor(),
                               SuperHeroLoader(),
                               transformer=SuperHeroReverseNameTransformer())

            job = DefaultJob(self.conf, task)
            job.launch()

            expected_list = ['{"hero": "Super man", "name": "tneK kralC"}',
                             '{"hero": "Bat man", "name": "enyaW ecurB"}']
            with open(self.dest_file_name, 'r') as file:
                for expected in expected_list:
                    actual = file.readline().rstrip('\n')
                    self.assertEqual(expected, actual)
                self.assertFalse(file.readline())

        self.assertEqual(mock_statsd.call_count, 0)


class TestJobNoTransform(unittest.TestCase):

    def setUp(self) -> None:
        self.temp_dir_path = tempfile.mkdtemp()
        self.dest_file_name = f'{self.temp_dir_path}/superhero.json'
        self.conf = ConfigFactory.from_dict(
            {'loader.superhero.dest_file': self.dest_file_name})

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir_path)

    def test_job(self) -> None:
        task = DefaultTask(SuperHeroExtractor(), SuperHeroLoader())

        job = DefaultJob(self.conf, task)
        job.launch()

        expected_list = ['{"hero": "Super man", "name": "Clark Kent"}',
                         '{"hero": "Bat man", "name": "Bruce Wayne"}']
        with open(self.dest_file_name, 'r') as file:
            for expected in expected_list:
                actual = file.readline().rstrip('\n')
                self.assertEqual(expected, actual)
            self.assertFalse(file.readline())


class TestJobStatsd(unittest.TestCase):

    def setUp(self) -> None:
        self.temp_dir_path = tempfile.mkdtemp()
        self.dest_file_name = f'{self.temp_dir_path}/superhero.json'
        self.conf = ConfigFactory.from_dict(
            {'loader.superhero.dest_file': self.dest_file_name,
             'job.is_statsd_enabled': True,
             'job.identifier': 'foobar'})

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir_path)

    def test_job(self) -> None:
        with patch("databuilder.job.job.StatsClient") as mock_statsd:
            task = DefaultTask(SuperHeroExtractor(), SuperHeroLoader())

            job = DefaultJob(self.conf, task)
            job.launch()

            expected_list = ['{"hero": "Super man", "name": "Clark Kent"}',
                             '{"hero": "Bat man", "name": "Bruce Wayne"}']
            with open(self.dest_file_name, 'r') as file:
                for expected in expected_list:
                    actual = file.readline().rstrip('\n')
                    self.assertEqual(expected, actual)
                self.assertFalse(file.readline())

            self.assertEqual(mock_statsd.return_value.incr.call_count, 1)


class SuperHeroExtractor(Extractor):
    def __init__(self) -> None:
        pass

    def init(self, conf: ConfigTree) -> None:
        self.records = [SuperHero(hero='Super man', name='Clark Kent'),
                        SuperHero(hero='Bat man', name='Bruce Wayne')]
        self.iter = iter(self.records)

    def extract(self) -> Any:
        try:
            return next(self.iter)
        except StopIteration:
            return None

    def get_scope(self) -> str:
        return 'extractor.superhero'


class SuperHero:
    def __init__(self,
                 hero: str,
                 name: str) -> None:
        self.hero = hero
        self.name = name

    def __repr__(self) -> str:
        return f'SuperHero(hero={self.hero}, name={self.name})'


class SuperHeroReverseNameTransformer(Transformer):
    def __init__(self) -> None:
        pass

    def init(self, conf: ConfigTree) -> None:
        pass

    def transform(self, record: Any) -> Any:
        record.name = record.name[::-1]
        return record

    def get_scope(self) -> str:
        return 'transformer.superhero'


class SuperHeroLoader(Loader):
    def init(self, conf: ConfigTree) -> None:
        self.conf = conf
        dest_file_path = self.conf.get_string('dest_file')
        LOGGER.info('Loading to %s', dest_file_path)
        self.dest_file_obj = open(self.conf.get_string('dest_file'), 'w')

    def load(self, record: Any) -> None:
        rec = json.dumps(record.__dict__, sort_keys=True)
        LOGGER.info('Writing record: %s', rec)
        self.dest_file_obj.write(f'{rec}\n')
        self.dest_file_obj.flush()

    def get_scope(self) -> str:
        return 'loader.superhero'


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from pyhocon import ConfigFactory

from databuilder.extractor.restapi.rest_api_extractor import (
    MODEL_CLASS, REST_API_QUERY, STATIC_RECORD_DICT, RestAPIExtractor,
)
from databuilder.models.dashboard.dashboard_metadata import DashboardMetadata
from databuilder.rest_api.base_rest_api_query import RestApiQuerySeed


class TestRestAPIExtractor(unittest.TestCase):

    def test_static_data(self) -> None:

        conf = ConfigFactory.from_dict(
            {
                REST_API_QUERY: RestApiQuerySeed(seed_record=[{'foo': 'bar'}]),
                STATIC_RECORD_DICT: {'john': 'doe'}
            }
        )
        extractor = RestAPIExtractor()
        extractor.init(conf=conf)

        record = extractor.extract()
        expected = {'foo': 'bar', 'john': 'doe'}

        self.assertDictEqual(expected, record)

    def test_model_construction(self) -> None:
        conf = ConfigFactory.from_dict(
            {
                REST_API_QUERY: RestApiQuerySeed(
                    seed_record=[{'dashboard_group': 'foo',
                                  'dashboard_name': 'bar',
                                  'description': 'john',
                                  'dashboard_group_description': 'doe'}]),
                MODEL_CLASS: 'databuilder.models.dashboard.dashboard_metadata.DashboardMetadata',
            }
        )
        extractor = RestAPIExtractor()
        extractor.init(conf=conf)

        record = extractor.extract()
        expected = DashboardMetadata(dashboard_group='foo', dashboard_name='bar', description='john',
                                     dashboard_group_description='doe')

        self.assertEqual(expected.__repr__(), record.__repr__())



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

"""
Shows how to set up a job that uses the neptune neptune staleness removal task
"""
import logging
import os

from pyhocon import ConfigFactory

from databuilder.clients.neptune_client import NeptuneSessionClient
from databuilder.job.job import DefaultJob
from databuilder.task.neptune_staleness_removal_task import NeptuneStalenessRemovalTask


def create_remove_stale_data_job():
    logging.basicConfig(level=logging.INFO)
    access_key = os.getenv('AWS_KEY')
    access_secret = os.getenv('AWS_SECRET_KEY')
    aws_zone = os.getenv("AWS_ZONE")
    neptune_host = os.getenv('CREDENTIALS_NEPTUNE_HOST', 'localhost')
    neptune_port = os.getenv('CREDENTIALS_NEPTUNE_PORT', 7687)
    neptune_host = '{}:{}'.format(neptune_host, neptune_port)
    target_relations = ['DESCRIPTION', 'DESCRIPTION_OF', 'COLUMN', 'COLUMN_OF', 'TABLE', 'TABLE_OF']
    target_nodes = ['Table', 'Column', 'Programmatic_Description', "Schema"]
    job_config = ConfigFactory.from_dict({
        'task.remove_stale_data': {
            NeptuneStalenessRemovalTask.TARGET_RELATIONS: target_relations,
            NeptuneStalenessRemovalTask.TARGET_NODES: target_nodes,
            NeptuneStalenessRemovalTask.STALENESS_CUT_OFF_IN_SECONDS: 86400,  # 1 day
            'neptune.client': {
                NeptuneSessionClient.NEPTUNE_HOST_NAME: neptune_host,
                NeptuneSessionClient.AWS_REGION: aws_zone,
                NeptuneSessionClient.AWS_ACCESS_KEY: access_key,
                NeptuneSessionClient.AWS_SECRET_ACCESS_KEY: access_secret
            }
        }
    })
    job = DefaultJob(
        conf=job_config,
        task=NeptuneStalenessRemovalTask()
    )
    return job


if __name__ == '__main__':
    job = create_remove_stale_data_job()
    job.launch()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import ANY

from databuilder.models.graph_serializable import (
    RELATION_END_KEY, RELATION_END_LABEL, RELATION_REVERSE_TYPE, RELATION_START_KEY, RELATION_START_LABEL,
    RELATION_TYPE,
)
from databuilder.models.table_lineage import TableLineage
from databuilder.serializers import neo4_serializer, neptune_serializer
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)

DB = 'hive'
SCHEMA = 'base'
TABLE = 'test'
CLUSTER = 'default'


class TestTableLineage(unittest.TestCase):

    def setUp(self) -> None:
        super(TestTableLineage, self).setUp()
        self.table_lineage = TableLineage(db_name='hive',
                                          schema=SCHEMA,
                                          table_name=TABLE,
                                          cluster=CLUSTER,
                                          downstream_deps=['hive://default.test_schema/test_table1',
                                                           'hive://default.test_schema/test_table2'])

        self.start_key = f'{DB}://{CLUSTER}.{SCHEMA}/{TABLE}'
        self.end_key1 = f'{DB}://{CLUSTER}.test_schema/test_table1'
        self.end_key2 = f'{DB}://{CLUSTER}.test_schema/test_table2'

    def test_get_table_model_key(self) -> None:
        metadata = self.table_lineage.get_table_model_key(db=DB,
                                                          cluster=CLUSTER,
                                                          schema=SCHEMA,
                                                          table=TABLE)
        self.assertEqual(metadata, 'hive://default.base/test')

    def test_create_nodes(self) -> None:
        actual = []
        node = self.table_lineage.create_next_node()
        while node:
            serialized_node = neo4_serializer.serialize_node(node)
            actual.append(serialized_node)
            node = self.table_lineage.create_next_node()

        self.assertEqual(len(actual), 0)

    def test_create_relation(self) -> None:
        expected_relations = [
            {
                RELATION_START_KEY: self.start_key,
                RELATION_START_LABEL: 'Table',
                RELATION_END_KEY: self.end_key1,
                RELATION_END_LABEL: 'Table',
                RELATION_TYPE: TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                RELATION_REVERSE_TYPE: TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE
            },
            {
                RELATION_START_KEY: self.start_key,
                RELATION_START_LABEL: 'Table',
                RELATION_END_KEY: self.end_key2,
                RELATION_END_LABEL: 'Table',
                RELATION_TYPE: TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                RELATION_REVERSE_TYPE: TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE
            }
        ]

        actual = []
        relation = self.table_lineage.create_next_relation()
        while relation:
            serialized_relation = neo4_serializer.serialize_relationship(relation)
            actual.append(serialized_relation)
            relation = self.table_lineage.create_next_relation()

        self.assertEqual(actual, expected_relations)

    def test_create_relation_neptune(self) -> None:
        expected = [
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:' + self.start_key,
                        to_vertex_id='Table:' + self.end_key1,
                        label=TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:' + self.start_key,
                        to_vertex_id='Table:' + self.end_key1,
                        label=TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Table:' + self.start_key,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Table:' + self.end_key1,
                    NEPTUNE_HEADER_LABEL: TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:' + self.end_key1,
                        to_vertex_id='Table:' + self.start_key,
                        label=TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:' + self.end_key1,
                        to_vertex_id='Table:' + self.start_key,
                        label=TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Table:' + self.end_key1,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Table:' + self.start_key,
                    NEPTUNE_HEADER_LABEL: TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ],
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:' + self.start_key,
                        to_vertex_id='Table:' + self.end_key2,
                        label=TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:' + self.start_key,
                        to_vertex_id='Table:' + self.end_key2,
                        label=TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Table:' + self.start_key,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Table:' + self.end_key2,
                    NEPTUNE_HEADER_LABEL: TableLineage.ORIGIN_DEPENDENCY_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:' + self.end_key2,
                        to_vertex_id='Table:' + self.start_key,
                        label=TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id='Table:' + self.end_key2,
                        to_vertex_id='Table:' + self.start_key,
                        label=TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: 'Table:' + self.end_key2,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: 'Table:' + self.start_key,
                    NEPTUNE_HEADER_LABEL: TableLineage.DEPENDENCY_ORIGIN_RELATION_TYPE,
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ]
        ]

        actual = []
        relation = self.table_lineage.create_next_relation()
        while relation:
            serialized_relation = neptune_serializer.convert_relationship(relation)
            actual.append(serialized_relation)
            relation = self.table_lineage.create_next_relation()

        self.assertEqual(actual, expected)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from pyhocon import ConfigFactory

from databuilder.models.dashboard.dashboard_execution import DashboardExecution
from databuilder.transformer.dict_to_model import MODEL_CLASS, DictToModel


class TestDictToModel(unittest.TestCase):

    def test_conversion(self) -> None:

        transformer = DictToModel()
        config = ConfigFactory.from_dict({
            MODEL_CLASS: 'databuilder.models.dashboard.dashboard_execution.DashboardExecution',
        })
        transformer.init(conf=config)

        actual = transformer.transform(
            {
                'dashboard_group_id': 'foo',
                'dashboard_id': 'bar',
                'execution_timestamp': 123456789,
                'execution_state': 'succeed',
                'product': 'mode',
                'cluster': 'gold'
            }
        )

        self.assertTrue(isinstance(actual, DashboardExecution))
        self.assertEqual(actual.__repr__(), DashboardExecution(
            dashboard_group_id='foo',
            dashboard_id='bar',
            execution_timestamp=123456789,
            execution_state='succeed',
            product='mode',
            cluster='gold'
        ).__repr__())



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import unittest
from collections import OrderedDict
from typing import Any

from cassandra.metadata import ColumnMetadata as CassandraColumnMetadata
from mock import patch
from pyhocon import ConfigFactory

from databuilder.extractor.cassandra_extractor import CassandraExtractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


# patch whole class to avoid actually calling for boto3.client during tests
@patch('cassandra.cluster.Cluster.connect', lambda x: None)
class TestCassandraExtractor(unittest.TestCase):
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)

        self.default_conf = ConfigFactory.from_dict({})

    def test_extraction_with_empty_query_result(self) -> None:
        """
        Test Extraction with empty result from query
        """
        extractor = CassandraExtractor()
        extractor.init(self.default_conf)

        results = extractor.extract()
        self.assertEqual(results, None)

    @patch('databuilder.extractor.cassandra_extractor.CassandraExtractor._get_keyspaces')
    @patch('databuilder.extractor.cassandra_extractor.CassandraExtractor._get_tables')
    @patch('databuilder.extractor.cassandra_extractor.CassandraExtractor._get_columns')
    def test_extraction_with_default_conf(self,
                                          mock_columns: Any,
                                          mock_tables: Any,
                                          mock_keyspaces: Any
                                          ) -> None:
        mock_keyspaces.return_value = {'test_schema': None}
        mock_tables.return_value = {'test_table': None}
        columns_dict = OrderedDict()
        columns_dict['id'] = CassandraColumnMetadata(None, 'id', 'int')
        columns_dict['txt'] = CassandraColumnMetadata(None, 'txt', 'text')
        mock_columns.return_value = columns_dict

        extractor = CassandraExtractor()
        extractor.init(self.default_conf)
        actual = extractor.extract()
        expected = TableMetadata('cassandra', 'gold', 'test_schema', 'test_table', None,
                                 [ColumnMetadata('id', None, 'int', 0),
                                  ColumnMetadata('txt', None, 'text', 1)])
        self.assertEqual(expected.__repr__(), actual.__repr__())
        self.assertIsNone(extractor.extract())

    @patch('databuilder.extractor.cassandra_extractor.CassandraExtractor._get_keyspaces')
    @patch('databuilder.extractor.cassandra_extractor.CassandraExtractor._get_tables')
    @patch('databuilder.extractor.cassandra_extractor.CassandraExtractor._get_columns')
    def test_extraction_with_filter_conf(self,
                                         mock_columns: Any,
                                         mock_tables: Any,
                                         mock_keyspaces: Any
                                         ) -> None:
        mock_keyspaces.return_value = {'test_schema': None}
        mock_tables.return_value = {'test_table': None}
        columns_dict = OrderedDict()
        columns_dict['id'] = CassandraColumnMetadata(None, 'id', 'int')
        columns_dict['txt'] = CassandraColumnMetadata(None, 'txt', 'text')
        mock_columns.return_value = columns_dict

        def filter_function(k: str, t: str) -> bool:
            return False if 'test' in k or 'test' in t else False

        conf = ConfigFactory.from_dict({
            CassandraExtractor.FILTER_FUNCTION_KEY: filter_function
        })

        extractor = CassandraExtractor()
        extractor.init(conf)
        self.assertIsNone(extractor.extract())


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from mock import MagicMock
from pyhocon import ConfigFactory

from databuilder.loader.generic_loader import CALLBACK_FUNCTION, GenericLoader


class TestGenericLoader(unittest.TestCase):

    def test_loading(self) -> None:

        loader = GenericLoader()
        callback_func = MagicMock()
        loader.init(conf=ConfigFactory.from_dict({
            CALLBACK_FUNCTION: callback_func
        }))

        loader.load({'foo': 'bar'})
        loader.close()

        callback_func.assert_called_once()

    def test_none_loading(self) -> None:

        loader = GenericLoader()
        callback_func = MagicMock()
        loader.init(conf=ConfigFactory.from_dict({
            CALLBACK_FUNCTION: callback_func
        }))

        loader.load(None)
        loader.close()

        callback_func.assert_not_called()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

"""
This is a example script which demo how to load data
into Neo4j and Elasticsearch without using an Airflow DAG.

"""

import os
import sys
import textwrap
import uuid

from elasticsearch import Elasticsearch
from pyhocon import ConfigFactory
from sqlalchemy.ext.declarative import declarative_base

from databuilder.extractor.neo4j_extractor import Neo4jExtractor
from databuilder.extractor.neo4j_search_data_extractor import Neo4jSearchDataExtractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.extractor.vertica_metadata_extractor import VerticaMetadataExtractor
from databuilder.job.job import DefaultJob
from databuilder.loader.file_system_elasticsearch_json_loader import FSElasticsearchJSONLoader
from databuilder.loader.file_system_neo4j_csv_loader import FsNeo4jCSVLoader
from databuilder.publisher import neo4j_csv_publisher
from databuilder.publisher.elasticsearch_publisher import ElasticsearchPublisher
from databuilder.publisher.neo4j_csv_publisher import Neo4jCsvPublisher
from databuilder.task.task import DefaultTask
from databuilder.transformer.base_transformer import NoopTransformer

es_host = os.getenv('CREDENTIALS_ELASTICSEARCH_PROXY_HOST', 'localhost')
neo_host = os.getenv('CREDENTIALS_NEO4J_PROXY_HOST', 'localhost')

es_port = os.getenv('CREDENTIALS_ELASTICSEARCH_PROXY_PORT', 9200)
neo_port = os.getenv('CREDENTIALS_NEO4J_PROXY_PORT', 7687)

if len(sys.argv) > 1:
    es_host = sys.argv[1]
if len(sys.argv) > 2:
    neo_host = sys.argv[2]

es = Elasticsearch([
    {'host': es_host, 'port': es_port},
])

DB_FILE = '/tmp/test.db'
SQLITE_CONN_STRING = 'sqlite:////tmp/test.db'
Base = declarative_base()

NEO4J_ENDPOINT = 'bolt://{}:{}'.format(neo_host, neo_port)

neo4j_endpoint = NEO4J_ENDPOINT

neo4j_user = 'neo4j'
neo4j_password = 'test'


# specify vertica access credentials, host server, port (default 5433),
# database name (default 'vertica')
def connection_string():
    user = 'username'
    password = 'password'
    host = 'vertica-budget.host'
    port = '5433'
    db = 'vertica'
    return "vertica+vertica_python://%s:%s@%s:%s/%s" % (user, password, host, port, db)


# provide schemas to run extraction on (default 'public')
def run_vertica_job():
    where_clause_suffix = textwrap.dedent("""
        where c.table_schema = 'public'
    """)

    tmp_folder = '/var/tmp/amundsen/table_metadata'
    node_files_folder = '{tmp_folder}/nodes/'.format(tmp_folder=tmp_folder)
    relationship_files_folder = '{tmp_folder}/relationships/'.format(tmp_folder=tmp_folder)

    job_config = ConfigFactory.from_dict({
        'extractor.vertica_metadata.{}'.format(VerticaMetadataExtractor.WHERE_CLAUSE_SUFFIX_KEY):
            where_clause_suffix,
        'extractor.vertica_metadata.{}'.format(VerticaMetadataExtractor.USE_CATALOG_AS_CLUSTER_NAME):
            False,
        'extractor.vertica_metadata.{}'.format(VerticaMetadataExtractor.CLUSTER_KEY):
            'vertica_budget',
        'extractor.vertica_metadata.extractor.sqlalchemy.{}'.format(SQLAlchemyExtractor.CONN_STRING):
            connection_string(),
        'loader.filesystem_csv_neo4j.{}'.format(FsNeo4jCSVLoader.NODE_DIR_PATH):
            node_files_folder,
        'loader.filesystem_csv_neo4j.{}'.format(FsNeo4jCSVLoader.RELATION_DIR_PATH):
            relationship_files_folder,
        'publisher.neo4j.{}'.format(neo4j_csv_publisher.NODE_FILES_DIR):
            node_files_folder,
        'publisher.neo4j.{}'.format(neo4j_csv_publisher.RELATION_FILES_DIR):
            relationship_files_folder,
        'publisher.neo4j.{}'.format(neo4j_csv_publisher.NEO4J_END_POINT_KEY):
            neo4j_endpoint,
        'publisher.neo4j.{}'.format(neo4j_csv_publisher.NEO4J_USER):
            neo4j_user,
        'publisher.neo4j.{}'.format(neo4j_csv_publisher.NEO4J_PASSWORD):
            neo4j_password,
        'publisher.neo4j.{}'.format(neo4j_csv_publisher.JOB_PUBLISH_TAG):
            'unique_tag',  # should use unique tag here like {ds}
    })
    job = DefaultJob(conf=job_config,
                     task=DefaultTask(extractor=VerticaMetadataExtractor(), loader=FsNeo4jCSVLoader()),
                     publisher=Neo4jCsvPublisher())
    return job


def create_es_publisher_sample_job(elasticsearch_index_alias='table_search_index',
                                   elasticsearch_doc_type_key='table',
                                   model_name='databuilder.models.table_elasticsearch_document.TableESDocument',
                                   cypher_query=None,
                                   elasticsearch_mapping=None):
    """
    :param elasticsearch_index_alias:  alias for Elasticsearch used in
                                       amundsensearchlibrary/search_service/config.py as an index
    :param elasticsearch_doc_type_key: name the ElasticSearch index is prepended with. Defaults to `table` resulting in
                                       `table_search_index`
    :param model_name:                 the Databuilder model class used in transporting between Extractor and Loader
    :param cypher_query:               Query handed to the `Neo4jSearchDataExtractor` class, if None is given (default)
                                       it uses the `Table` query baked into the Extractor
    :param elasticsearch_mapping:      Elasticsearch field mapping "DDL" handed to the `ElasticsearchPublisher` class,
                                       if None is given (default) it uses the `Table` query baked into the Publisher
    """
    # loader saves data to this location and publisher reads it from here
    extracted_search_data_path = '/var/tmp/amundsen/search_data.json'

    task = DefaultTask(loader=FSElasticsearchJSONLoader(),
                       extractor=Neo4jSearchDataExtractor(),
                       transformer=NoopTransformer())

    # elastic search client instance
    elasticsearch_client = es
    # unique name of new index in Elasticsearch
    elasticsearch_new_index_key = 'tables' + str(uuid.uuid4())

    job_config = ConfigFactory.from_dict({
        'extractor.search_data.extractor.neo4j.{}'.format(Neo4jExtractor.GRAPH_URL_CONFIG_KEY): neo4j_endpoint,
        'extractor.search_data.extractor.neo4j.{}'.format(Neo4jExtractor.MODEL_CLASS_CONFIG_KEY): model_name,
        'extractor.search_data.extractor.neo4j.{}'.format(Neo4jExtractor.NEO4J_AUTH_USER): neo4j_user,
        'extractor.search_data.extractor.neo4j.{}'.format(Neo4jExtractor.NEO4J_AUTH_PW): neo4j_password,
        'loader.filesystem.elasticsearch.{}'.format(FSElasticsearchJSONLoader.FILE_PATH_CONFIG_KEY):
            extracted_search_data_path,
        'loader.filesystem.elasticsearch.{}'.format(FSElasticsearchJSONLoader.FILE_MODE_CONFIG_KEY): 'w',
        'publisher.elasticsearch.{}'.format(ElasticsearchPublisher.FILE_PATH_CONFIG_KEY):
            extracted_search_data_path,
        'publisher.elasticsearch.{}'.format(ElasticsearchPublisher.FILE_MODE_CONFIG_KEY): 'r',
        'publisher.elasticsearch.{}'.format(ElasticsearchPublisher.ELASTICSEARCH_CLIENT_CONFIG_KEY):
            elasticsearch_client,
        'publisher.elasticsearch.{}'.format(ElasticsearchPublisher.ELASTICSEARCH_NEW_INDEX_CONFIG_KEY):
            elasticsearch_new_index_key,
        'publisher.elasticsearch.{}'.format(ElasticsearchPublisher.ELASTICSEARCH_DOC_TYPE_CONFIG_KEY):
            elasticsearch_doc_type_key,
        'publisher.elasticsearch.{}'.format(ElasticsearchPublisher.ELASTICSEARCH_ALIAS_CONFIG_KEY):
            elasticsearch_index_alias,
    })

    # only optionally add these keys, so need to dynamically `put` them
    if cypher_query:
        job_config.put('extractor.search_data.{}'.format(Neo4jSearchDataExtractor.CYPHER_QUERY_CONFIG_KEY),
                       cypher_query)
    if elasticsearch_mapping:
        job_config.put('publisher.elasticsearch.{}'.format(ElasticsearchPublisher.ELASTICSEARCH_MAPPING_CONFIG_KEY),
                       elasticsearch_mapping)

    job = DefaultJob(conf=job_config,
                     task=task,
                     publisher=ElasticsearchPublisher())
    return job


if __name__ == "__main__":
    # Uncomment next line to get INFO level logging
    # logging.basicConfig(level=logging.INFO)

    loading_job = run_vertica_job()
    loading_job.launch()

    job_es_table = create_es_publisher_sample_job(
        elasticsearch_index_alias='table_search_index',
        elasticsearch_doc_type_key='table',
        model_name='databuilder.models.table_elasticsearch_document.TableESDocument')
    job_es_table.launch()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import collections
import csv
import logging
import os
import unittest
from operator import itemgetter
from os import listdir
from os.path import isfile, join
from typing import (
    Any, Callable, Dict, Iterable, Optional, Union,
)

from pyhocon import ConfigFactory, ConfigTree

from databuilder.job.base_job import Job
from databuilder.loader.file_system_neo4j_csv_loader import FsNeo4jCSVLoader
from databuilder.models.graph_serializable import (
    GraphNode, GraphRelationship, GraphSerializable,
)
from tests.unit.models.test_graph_serializable import (
    Actor, City, Movie,
)

here = os.path.dirname(__file__)


class TestFsNeo4jCSVLoader(unittest.TestCase):
    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)

    def tearDown(self) -> None:
        Job.closer.close()

    def test_load(self) -> None:
        actors = [Actor('Tom Cruise'), Actor('Meg Ryan')]
        cities = [City('San Diego'), City('Oakland')]
        movie = Movie('Top Gun', actors, cities)

        loader = FsNeo4jCSVLoader()

        folder = 'movies'
        conf = self._make_conf(folder)

        loader.init(conf)
        loader.load(movie)
        loader.close()

        expected_node_path = os.path.join(here, f'../resources/fs_neo4j_csv_loader/{folder}/nodes')
        expected_nodes = self._get_csv_rows(expected_node_path, itemgetter('KEY'))
        actual_nodes = self._get_csv_rows(conf.get_string(FsNeo4jCSVLoader.NODE_DIR_PATH),
                                          itemgetter('KEY'))
        self.assertEqual(expected_nodes, actual_nodes)

        expected_rel_path = os.path.join(here, f'../resources/fs_neo4j_csv_loader/{folder}/relationships')
        expected_relations = self._get_csv_rows(expected_rel_path, itemgetter('START_KEY', 'END_KEY'))
        actual_relations = self._get_csv_rows(conf.get_string(FsNeo4jCSVLoader.RELATION_DIR_PATH),
                                              itemgetter('START_KEY', 'END_KEY'))
        self.assertEqual(expected_relations, actual_relations)

    def test_load_disjoint_properties(self) -> None:
        people = [
            Person("Taylor", job="Engineer"),
            Person("Griffin", pet="Lion"),
        ]

        loader = FsNeo4jCSVLoader()

        folder = 'people'
        conf = self._make_conf(folder)

        loader.init(conf)
        loader.load(people[0])
        loader.load(people[1])
        loader.close()

        expected_node_path = os.path.join(here, f'../resources/fs_neo4j_csv_loader/{folder}/nodes')
        expected_nodes = self._get_csv_rows(expected_node_path, itemgetter('KEY'))
        actual_nodes = self._get_csv_rows(conf.get_string(FsNeo4jCSVLoader.NODE_DIR_PATH),
                                          itemgetter('KEY'))
        self.assertEqual(expected_nodes, actual_nodes)

    def _make_conf(self, test_name: str) -> ConfigTree:
        prefix = '/var/tmp/TestFsNeo4jCSVLoader'

        return ConfigFactory.from_dict({
            FsNeo4jCSVLoader.NODE_DIR_PATH: f'{prefix}/{test_name}/{"nodes"}',
            FsNeo4jCSVLoader.RELATION_DIR_PATH: f'{prefix}/{test_name}/{"relationships"}',
            FsNeo4jCSVLoader.SHOULD_DELETE_CREATED_DIR: True
        })

    def _get_csv_rows(self,
                      path: str,
                      sorting_key_getter: Callable) -> Iterable[Dict[str, Any]]:
        files = [join(path, f) for f in listdir(path) if isfile(join(path, f))]

        result = []
        for f in files:
            with open(f, 'r') as f_input:
                reader = csv.DictReader(f_input)
                for row in reader:
                    result.append(collections.OrderedDict(sorted(row.items())))

        return sorted(result, key=sorting_key_getter)


class Person(GraphSerializable):
    """ A Person has multiple optional attributes. When an attribute is None,
        it is not included in the resulting node.
    """
    LABEL = 'Person'
    KEY_FORMAT = 'person://{}'

    def __init__(self,
                 name: str,
                 *,
                 pet: Optional[str] = None,
                 job: Optional[str] = None,
                 ) -> None:
        self._name = name
        self._pet = pet
        self._job = job
        self._node_iter = iter(self.create_nodes())

    def create_next_node(self) -> Union[GraphNode, None]:
        try:
            return next(self._node_iter)
        except StopIteration:
            return None

    def create_next_relation(self) -> Union[GraphRelationship, None]:
        return None

    def create_nodes(self) -> Iterable[GraphNode]:
        attributes = {"name": self._name}
        if self._pet:
            attributes['pet'] = self._pet
        if self._job:
            attributes['job'] = self._job

        return [GraphNode(
            key=Person.KEY_FORMAT.format(self._name),
            label=Person.LABEL,
            attributes=attributes
        )]


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import logging
import unittest
from typing import (
    Any, Dict, List,
)
from unittest.mock import MagicMock, patch

from pyhocon import ConfigFactory

from databuilder.extractor.dremio_metadata_extractor import DremioMetadataExtractor
from databuilder.models.table_metadata import ColumnMetadata, TableMetadata


class TestDremioMetadataExtractor(unittest.TestCase):

    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)

        config_dict: Dict[str, str] = {}

        self.conf = ConfigFactory.from_dict(config_dict)

    @patch('databuilder.extractor.dremio_metadata_extractor.connect')
    def test_extraction_with_empty_query_result(self, mock_connect: MagicMock) -> None:
        """
        Test Extraction with empty result from query
        """
        mock_connection = MagicMock()
        mock_connect.return_value = mock_connection

        mock_cursor = MagicMock()
        mock_connection.cursor.return_value = mock_cursor

        extractor = DremioMetadataExtractor()
        extractor.init(self.conf)

        results = extractor.extract()
        self.assertEqual(results, None)

    @patch('databuilder.extractor.dremio_metadata_extractor.connect')
    def test_extraction_with_single_result(self, mock_connect: MagicMock) -> None:
        """
        Test Extraction with single table result from query
        """
        mock_connection = MagicMock()
        mock_connect.return_value = mock_connection

        mock_cursor = MagicMock()
        mock_connection.cursor.return_value = mock_cursor

        mock_execute = MagicMock()
        mock_cursor.execute = mock_execute

        mock_cursor.description = [
            ['col_name'],
            ['col_description'],
            ['col_type'],
            ['col_sort_order'],
            ['database'],
            ['cluster'],
            ['schema'],
            ['name'],
            ['description'],
            ['is_view']
        ]

        # Pass flake8 Unsupported operand types for + error
        table: List[Any] = [
            'DREMIO',
            'Production',
            'test_schema',
            'test_table',
            'a table for testing',
            'false'
        ]

        # Pass flake8 Unsupported operand types for + error
        expected_input: List[List[Any]] = [
            ['col_id1', 'description of id1', 'number', 0] + table,
            ['col_id2', 'description of id2', 'number', 1] + table,
            ['is_active', None, 'boolean', 2] + table,
            ['source', 'description of source', 'varchar', 3] + table,
            ['etl_created_at', 'description of etl_created_at', 'timestamp_ltz', 4] + table,
            ['ds', None, 'varchar', 5] + table
        ]

        mock_cursor.execute.return_value = expected_input

        extractor = DremioMetadataExtractor()
        extractor.init(self.conf)

        actual = extractor.extract()
        expected = TableMetadata('DREMIO', 'Production', 'test_schema', 'test_table', 'a table for testing',
                                 [ColumnMetadata('col_id1', 'description of id1', 'number', 0),
                                  ColumnMetadata('col_id2', 'description of id2', 'number', 1),
                                  ColumnMetadata('is_active', None, 'boolean', 2),
                                  ColumnMetadata('source', 'description of source', 'varchar', 3),
                                  ColumnMetadata('etl_created_at', 'description of etl_created_at',
                                                 'timestamp_ltz', 4),
                                  ColumnMetadata('ds', None, 'varchar', 5)])

        self.assertEqual(expected.__repr__(), actual.__repr__())
        self.assertIsNone(extractor.extract())


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import ANY

from databuilder.models.graph_node import GraphNode
from databuilder.models.graph_relationship import GraphRelationship
from databuilder.models.graph_serializable import (
    NODE_KEY, NODE_LABEL, RELATION_END_KEY, RELATION_END_LABEL, RELATION_REVERSE_TYPE, RELATION_START_KEY,
    RELATION_START_LABEL, RELATION_TYPE,
)
from databuilder.models.watermark import Watermark
from databuilder.serializers import (
    mysql_serializer, neo4_serializer, neptune_serializer,
)
from databuilder.serializers.neptune_serializer import (
    METADATA_KEY_PROPERTY_NAME, NEPTUNE_CREATION_TYPE_JOB, NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT,
    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_HEADER_ID, NEPTUNE_HEADER_LABEL,
    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT, NEPTUNE_RELATIONSHIP_HEADER_FROM,
    NEPTUNE_RELATIONSHIP_HEADER_TO,
)

CREATE_TIME = '2017-09-18T00:00:00'
DATABASE = 'DYNAMO'
SCHEMA = 'BASE'
TABLE = 'TEST'
NESTED_PART = 'ds=2017-09-18/feature_id=9'
CLUSTER = 'DEFAULT'
PART_TYPE = 'LOW_WATERMARK'


class TestWatermark(unittest.TestCase):

    def setUp(self) -> None:
        super(TestWatermark, self).setUp()
        self.watermark = Watermark(
            create_time='2017-09-18T00:00:00',
            database=DATABASE,
            schema=SCHEMA,
            table_name=TABLE,
            cluster=CLUSTER,
            part_type=PART_TYPE,
            part_name=NESTED_PART
        )
        self.start_key = f'{DATABASE}://{CLUSTER}.{SCHEMA}/{TABLE}/{PART_TYPE}/'
        self.end_key = f'{DATABASE}://{CLUSTER}.{SCHEMA}/{TABLE}'
        self.expected_node_result = GraphNode(
            key=self.start_key,
            label='Watermark',
            attributes={
                'partition_key': 'ds',
                'partition_value': '2017-09-18/feature_id=9',
                'create_time': '2017-09-18T00:00:00'
            }
        )

        self.expected_serialized_node_results = [{
            NODE_KEY: self.start_key,
            NODE_LABEL: 'Watermark',
            'partition_key': 'ds',
            'partition_value': '2017-09-18/feature_id=9',
            'create_time': '2017-09-18T00:00:00'
        }]

        self.expected_relation_result = GraphRelationship(
            start_label='Watermark',
            end_label='Table',
            start_key=self.start_key,
            end_key=self.end_key,
            type='BELONG_TO_TABLE',
            reverse_type='WATERMARK',
            attributes={}
        )

        self.expected_serialized_relation_results = [{
            RELATION_START_KEY: self.start_key,
            RELATION_START_LABEL: 'Watermark',
            RELATION_END_KEY: self.end_key,
            RELATION_END_LABEL: 'Table',
            RELATION_TYPE: 'BELONG_TO_TABLE',
            RELATION_REVERSE_TYPE: 'WATERMARK'
        }]

    def test_get_watermark_model_key(self) -> None:
        watermark = self.watermark.get_watermark_model_key()
        self.assertEqual(watermark, f'{DATABASE}://{CLUSTER}.{SCHEMA}/{TABLE}/{PART_TYPE}/')

    def test_get_metadata_model_key(self) -> None:
        metadata = self.watermark.get_metadata_model_key()
        self.assertEqual(metadata, f'{DATABASE}://{CLUSTER}.{SCHEMA}/{TABLE}')

    def test_create_nodes(self) -> None:
        actual = []
        node = self.watermark.create_next_node()
        while node:
            serialized_node = neo4_serializer.serialize_node(node)
            actual.append(serialized_node)
            node = self.watermark.create_next_node()

        self.assertEqual(actual, self.expected_serialized_node_results)

    def test_create_nodes_neptune(self) -> None:
        expected_serialized_node_results = [{
            NEPTUNE_HEADER_ID: 'Watermark:' + self.start_key,
            METADATA_KEY_PROPERTY_NAME: 'Watermark:' + self.start_key,
            NEPTUNE_HEADER_LABEL: 'Watermark',
            NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
            NEPTUNE_CREATION_TYPE_NODE_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB,
            'partition_key:String(single)': 'ds',
            'partition_value:String(single)': '2017-09-18/feature_id=9',
            'create_time:String(single)': '2017-09-18T00:00:00'
        }]

        actual = []
        node = self.watermark.create_next_node()
        while node:
            serialized_node = neptune_serializer.convert_node(node)
            actual.append(serialized_node)
            node = self.watermark.create_next_node()

        self.assertEqual(expected_serialized_node_results, actual)

    def test_create_relation(self) -> None:
        actual = []
        relation = self.watermark.create_next_relation()
        while relation:
            serialized_relation = neo4_serializer.serialize_relationship(relation)
            actual.append(serialized_relation)
            relation = self.watermark.create_next_relation()

        self.assertEqual(actual, self.expected_serialized_relation_results)

    def test_create_relation_neptune(self) -> None:
        actual = []
        relation = self.watermark.create_next_relation()
        while relation:
            serialized_relation = neptune_serializer.convert_relationship(relation)
            actual.append(serialized_relation)
            relation = self.watermark.create_next_relation()

        expected = [
            [
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Watermark:" + self.start_key,
                        to_vertex_id="Table:" + self.end_key,
                        label='BELONG_TO_TABLE'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Watermark:" + self.start_key,
                        to_vertex_id="Table:" + self.end_key,
                        label='BELONG_TO_TABLE'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: "Watermark:" + self.start_key,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: "Table:" + self.end_key,
                    NEPTUNE_HEADER_LABEL: 'BELONG_TO_TABLE',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                },
                {
                    NEPTUNE_HEADER_ID: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Table:" + self.end_key,
                        to_vertex_id="Watermark:" + self.start_key,
                        label='WATERMARK'
                    ),
                    METADATA_KEY_PROPERTY_NAME: "{label}:{from_vertex_id}_{to_vertex_id}".format(
                        from_vertex_id="Table:" + self.end_key,
                        to_vertex_id="Watermark:" + self.start_key,
                        label='WATERMARK'
                    ),
                    NEPTUNE_RELATIONSHIP_HEADER_FROM: "Table:" + self.end_key,
                    NEPTUNE_RELATIONSHIP_HEADER_TO: "Watermark:" + self.start_key,
                    NEPTUNE_HEADER_LABEL: 'WATERMARK',
                    NEPTUNE_LAST_EXTRACTED_AT_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: ANY,
                    NEPTUNE_CREATION_TYPE_RELATIONSHIP_PROPERTY_NAME_BULK_LOADER_FORMAT: NEPTUNE_CREATION_TYPE_JOB
                }
            ]
        ]

        self.assertListEqual(actual, expected)

    def test_create_records(self) -> None:
        expected = [{
            'rk': self.start_key,
            'partition_key': 'ds',
            'partition_value': '2017-09-18/feature_id=9',
            'create_time': '2017-09-18T00:00:00',
            'table_rk': self.end_key
        }]

        actual = []
        record = self.watermark.create_next_record()
        while record:
            serialized_record = mysql_serializer.serialize_record(record)
            actual.append(serialized_record)
            record = self.watermark.create_next_record()

        self.assertEqual(actual, expected)



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import itertools
import logging
import unittest
from datetime import datetime
from typing import (
    Iterable, Iterator, Optional, TypeVar,
)

from mock import MagicMock, patch
from pyhocon import ConfigFactory
from pytz import UTC

from databuilder.extractor.hive_table_last_updated_extractor import HiveTableLastUpdatedExtractor
from databuilder.extractor.sql_alchemy_extractor import SQLAlchemyExtractor
from databuilder.filesystem.filesystem import FileSystem
from databuilder.filesystem.metadata import FileMetadata
from databuilder.models.table_last_updated import TableLastUpdated

T = TypeVar('T')


def null_iterator(items: Iterable[T]) -> Iterator[Optional[T]]:
    """
    Returns an infinite iterator that returns the items from items,
    then infinite Nones. Required because Extractor.extract is expected
    to return None when it is exhausted, not terminate.
    """
    return itertools.chain(iter(items), itertools.repeat(None))


class TestHiveTableLastUpdatedExtractor(unittest.TestCase):

    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)

    def test_extraction_with_empty_query_result(self) -> None:
        config_dict = {
            f'extractor.sqlalchemy.{SQLAlchemyExtractor.CONN_STRING}': 'TEST_CONNECTION',
            f'filesystem.{FileSystem.DASK_FILE_SYSTEM}': MagicMock()
        }
        conf = ConfigFactory.from_dict(config_dict)
        with patch.object(SQLAlchemyExtractor, '_get_connection'):
            extractor = HiveTableLastUpdatedExtractor()
            extractor.init(conf)

            result = extractor.extract()
            self.assertEqual(result, None)

    def test_extraction_with_partition_table_result(self) -> None:
        config_dict = {
            f'filesystem.{FileSystem.DASK_FILE_SYSTEM}': MagicMock()
        }
        conf = ConfigFactory.from_dict(config_dict)

        pt_alchemy_extractor_instance = MagicMock()
        non_pt_alchemy_extractor_instance = MagicMock()
        with patch.object(HiveTableLastUpdatedExtractor, '_get_partitioned_table_sql_alchemy_extractor',
                          return_value=pt_alchemy_extractor_instance), \
            patch.object(HiveTableLastUpdatedExtractor, '_get_non_partitioned_table_sql_alchemy_extractor',
                         return_value=non_pt_alchemy_extractor_instance):
            pt_alchemy_extractor_instance.extract = MagicMock(side_effect=null_iterator([{
                'schema': 'foo_schema',
                'table_name': 'table_1',
                'last_updated_time': 1
            }, {
                'schema': 'foo_schema',
                'table_name': 'table_2',
                'last_updated_time': 2
            }]))

            non_pt_alchemy_extractor_instance.extract = MagicMock(return_value=None)

            extractor = HiveTableLastUpdatedExtractor()
            extractor.init(conf)

            result = extractor.extract()
            expected = TableLastUpdated(schema='foo_schema', table_name='table_1', last_updated_time_epoch=1,
                                        db='hive', cluster='gold')
            self.assertEqual(result.__repr__(), expected.__repr__())
            result = extractor.extract()
            expected = TableLastUpdated(schema='foo_schema', table_name='table_2', last_updated_time_epoch=2,
                                        db='hive', cluster='gold')
            self.assertEqual(result.__repr__(), expected.__repr__())

            self.assertIsNone(extractor.extract())

    def test_extraction(self) -> None:
        old_datetime = datetime(2018, 8, 14, 4, 12, 3, tzinfo=UTC)
        new_datetime = datetime(2018, 11, 14, 4, 12, 3, tzinfo=UTC)

        fs = MagicMock()
        fs.ls = MagicMock(return_value=['/foo/bar', '/foo/baz'])
        fs.is_file = MagicMock(return_value=True)
        fs.info = MagicMock(side_effect=[
            FileMetadata(path='/foo/bar', last_updated=old_datetime, size=15093),
            FileMetadata(path='/foo/baz', last_updated=new_datetime, size=15094)
        ])

        pt_alchemy_extractor_instance = MagicMock()
        non_pt_alchemy_extractor_instance = MagicMock()

        with patch.object(HiveTableLastUpdatedExtractor, '_get_partitioned_table_sql_alchemy_extractor',
                          return_value=pt_alchemy_extractor_instance), \
            patch.object(HiveTableLastUpdatedExtractor, '_get_non_partitioned_table_sql_alchemy_extractor',
                         return_value=non_pt_alchemy_extractor_instance), \
            patch.object(HiveTableLastUpdatedExtractor, '_get_filesystem',
                         return_value=fs):
            pt_alchemy_extractor_instance.extract = MagicMock(return_value=None)

            non_pt_alchemy_extractor_instance.extract = MagicMock(side_effect=null_iterator([{
                'schema': 'foo_schema',
                'table_name': 'table_1',
                'location': '/foo/bar'
            }]))

            extractor = HiveTableLastUpdatedExtractor()
            extractor.init(ConfigFactory.from_dict({}))

            result = extractor.extract()
            expected = TableLastUpdated(schema='foo_schema', table_name='table_1',
                                        last_updated_time_epoch=1542168723,
                                        db='hive', cluster='gold')
            self.assertEqual(result.__repr__(), expected.__repr__())

            self.assertIsNone(extractor.extract())


if __name__ == '__main__':
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import json
import re
import unittest

from feast.entity import Entity
from feast.feature_table import FeatureTable
from mock import MagicMock, call
from pyhocon import ConfigFactory

from databuilder import Scoped
from databuilder.extractor.feast_extractor import FeastExtractor
from databuilder.models.table_metadata import (
    ColumnMetadata, DescriptionMetadata, TableMetadata,
)


class TestFeastExtractor(unittest.TestCase):
    def test_no_feature_tables_registered(self) -> None:
        self._init_extractor()
        self.extractor._client.list_projects.return_value = ["default"]

        self.assertIsNone(self.extractor.extract())

    def test_every_project_is_scanned(self) -> None:
        self._init_extractor()
        self.extractor._client.list_projects.return_value = ["default", "dev", "prod"]
        list_feature_tables_mock = self.extractor._client.list_feature_tables
        list_feature_tables_mock.return_value = []

        self.assertIsNone(self.extractor.extract())
        list_feature_tables_mock.assert_has_calls(
            [
                call(project="default"),
                call(project="dev"),
                call(project="prod"),
            ]
        )

    def test_feature_table_extraction(self) -> None:
        self._init_extractor(programmatic_description_enabled=False)
        self.extractor._client.list_projects.return_value = ["default"]
        self._mock_feature_table()

        table = self.extractor.extract()
        self.extractor._client.get_entity.assert_called_with(
            "driver_id", project="default"
        )
        expected = TableMetadata(
            database="feast",
            cluster="unittest-feast-instance",
            schema="default",
            name="driver_trips",
            description=None,
            columns=[
                ColumnMetadata(
                    "driver_id", "Internal identifier of the driver", "INT64", 0
                ),
                ColumnMetadata("trips_today", None, "INT32", 1),
            ],
        )

        self.assertEqual(expected.__repr__(), table.__repr__())
        self.assertIsNone(self.extractor.extract())

    def test_feature_table_extraction_with_description_batch(self) -> None:
        self._init_extractor(programmatic_description_enabled=True)
        self.extractor._client.list_projects.return_value = ["default"]
        self._mock_feature_table(labels={"label1": "value1"})

        feature_table_definition = self.extractor.extract()
        assert isinstance(feature_table_definition, TableMetadata)

        description = self.extractor.extract()
        assert isinstance(description, TableMetadata)
        expected = DescriptionMetadata(
            TestFeastExtractor._strip_margin(
                """* Created at **2020-01-01 00:00:00**
                  |* Labels:
                  |    * label1: **value1**
                  |"""
            ),
            "feature_table_details",
        )
        self.assertEqual(expected.__repr__(), description.description.__repr__())

        batch_source = self.extractor.extract()
        assert isinstance(batch_source, TableMetadata)
        expected = DescriptionMetadata(
            TestFeastExtractor._strip_margin(
                """```
                |fileOptions:
                |  fileFormat:
                |    parquetFormat: {}
                |  fileUrl: file:///some/location
                |type: BATCH_FILE
                |```"""
            ),
            "batch_source",
        )
        self.assertEqual(expected.__repr__(), batch_source.description.__repr__())

        self.assertIsNone(self.extractor.extract())

    def test_feature_table_extraction_with_description_stream(self) -> None:
        self._init_extractor(programmatic_description_enabled=True)
        self.extractor._client.list_projects.return_value = ["default"]
        self._mock_feature_table(add_stream_source=True)

        feature_table_definition = self.extractor.extract()
        assert isinstance(feature_table_definition, TableMetadata)

        description = self.extractor.extract()
        assert isinstance(description, TableMetadata)
        expected = DescriptionMetadata(
            TestFeastExtractor._strip_margin(
                """* Created at **2020-01-01 00:00:00**
                  |"""
            ),
            "feature_table_details",
        )
        self.assertEqual(expected.__repr__(), description.description.__repr__())

        batch_source = self.extractor.extract()
        assert isinstance(batch_source, TableMetadata)
        expected = DescriptionMetadata(
            TestFeastExtractor._strip_margin(
                """```
                |fileOptions:
                |  fileFormat:
                |    parquetFormat: {}
                |  fileUrl: file:///some/location
                |type: BATCH_FILE
                |```"""
            ),
            "batch_source",
        )
        self.assertEqual(expected.__repr__(), batch_source.description.__repr__())

        stream_source = self.extractor.extract()
        assert isinstance(stream_source, TableMetadata)
        expected = DescriptionMetadata(
            TestFeastExtractor._strip_margin(
                """```
                 |createdTimestampColumn: datetime
                 |eventTimestampColumn: datetime
                 |kafkaOptions:
                 |  bootstrapServers: broker1
                 |  messageFormat:
                 |    avroFormat:
                 |      schemaJson: '{"type": "record", "name": "DriverTrips", "fields": [{"name": "driver_id",
                 |        "type": "long"}, {"name": "trips_today", "type": "int"}, {"name": "datetime",
                 |        "type": {"type": "long", "logicalType": "timestamp-micros"}}]}'
                 |  topic: driver_trips
                 |type: STREAM_KAFKA
                 |```"""
            ),
            "stream_source",
        )
        self.assertEqual(expected.__repr__(), stream_source.description.__repr__())

        self.assertIsNone(self.extractor.extract())

    def _init_extractor(self, programmatic_description_enabled: bool = True) -> None:
        conf = {
            f'extractor.feast.{FeastExtractor.FEAST_ENDPOINT_CONFIG_KEY}': 'feast-core.example.com:6565',
            f'extractor.feast.{FeastExtractor.FEAST_SERVICE_CONFIG_KEY}': 'unittest-feast-instance',
            f'extractor.feast.{FeastExtractor.DESCRIBE_FEATURE_TABLES}': programmatic_description_enabled,
        }
        self.extractor = FeastExtractor()
        self.extractor.init(
            Scoped.get_scoped_conf(
                conf=ConfigFactory.from_dict(conf), scope=self.extractor.get_scope()
            )
        )
        self.extractor._client = MagicMock(return_value=None)

    @staticmethod
    def _strip_margin(text: str) -> str:
        return re.sub("\n[ \t]*\\|", "\n", text)

    def _mock_feature_table(
            self, labels: dict = {}, add_stream_source: bool = False
    ) -> None:
        table_spec = {
            "name": "driver_trips",
            "entities": ["driver_id"],
            "features": [{"name": "trips_today", "valueType": "INT32"}],
            "labels": labels,
            "batchSource": {
                "type": "BATCH_FILE",
                "fileOptions": {
                    "fileFormat": {"parquetFormat": {}},
                    "fileUrl": "file:///some/location",
                },
            },
        }

        if add_stream_source:
            avro_schema_json = json.dumps(
                {
                    "type": "record",
                    "name": "DriverTrips",
                    "fields": [
                        {"name": "driver_id", "type": "long"},
                        {"name": "trips_today", "type": "int"},
                        {
                            "name": "datetime",
                            "type": {"type": "long", "logicalType": "timestamp-micros"},
                        },
                    ],
                }
            )

            table_spec["streamSource"] = {
                "type": "STREAM_KAFKA",
                "eventTimestampColumn": "datetime",
                "createdTimestampColumn": "datetime",
                "kafkaOptions": {
                    "bootstrapServers": "broker1",
                    "topic": "driver_trips",
                    "messageFormat": {
                        "avroFormat": {
                            "schemaJson": avro_schema_json,
                        }
                    },
                },
            }

        self.extractor._client.list_feature_tables.return_value = [
            FeatureTable.from_dict(
                {
                    "spec": table_spec,
                    "meta": {"createdTimestamp": "2020-01-01T00:00:00Z"},
                }
            )
        ]
        self.extractor._client.get_entity.return_value = Entity.from_dict(
            {
                "spec": {
                    "name": "driver_id",
                    "valueType": "INT64",
                    "description": "Internal identifier of the driver",
                }
            }
        )


if __name__ == "__main__":
    unittest.main()



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import unittest

from mock import MagicMock

from databuilder.rest_api.rest_api_failure_handlers import HttpFailureSkipOnStatus


class TestHttpFailureSkipOnStatus(unittest.TestCase):

    def testSkip(self) -> None:
        failure_handler = HttpFailureSkipOnStatus([404, 400])

        exception = MagicMock()
        exception.response.status_code = 404
        self.assertTrue(failure_handler.can_skip_failure(exception=exception))

        exception.response.status_code = 400
        self.assertTrue(failure_handler.can_skip_failure(exception=exception))

        exception.response.status_code = 500
        self.assertFalse(failure_handler.can_skip_failure(exception=exception))



# Copyright Contributors to the Amundsen project.
# SPDX-License-Identifier: Apache-2.0

import io
import os
import unittest

import responses
from pyhocon import ConfigFactory

from databuilder.extractor.user.bamboohr.bamboohr_user_extractor import BamboohrUserExtractor
from databuilder.models.user import User


class TestBamboohrUserExtractor(unittest.TestCase):
    @responses.activate
    def test_parse_testdata(self) -> None:
        bhr = BamboohrUserExtractor()
        bhr.init(ConfigFactory.from_dict({'api_key': 'api_key', 'subdomain': 'amundsen'}))

        testdata_xml = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            '../../../resources/extractor/user/bamboohr/testdata.xml'
        )

        with io.open(testdata_xml) as testdata:
            responses.add(responses.GET, bhr._employee_directory_uri(), body=testdata.read())

        expected = User(
            email='roald@amundsen.io',
            first_name='Roald',
            last_name='Amundsen',
            name='Roald Amundsen',
            team_name='508 Corporate Marketing',
            role_name='Antarctic Explorer',
        )

        actual_users = list(bhr._get_extract_iter())

        self.assertEqual(1, len(actual_users))
        self.assertEqual(repr(expected), repr(actual_users[0]))


if __name__ == '__main__':
    unittest.main()
